import warnings
import logging
from pathlib import Path

import geopandas
import pandas
import pandas as pd
import geopandas as gpd

from src.agents.firm import Firm, Firms
from src.network.mrio import Mrio
from src.model.builder_functions import get_index_closest_point, get_closest_road_nodes, get_long_lat


def create_firms(
        firm_table: pd.DataFrame,
        keep_top_n_firms: object = None,
        inventory_restoration_time: float = 4,
        utilization_rate: float = 0.8,
        capital_to_value_added_ratio: float = 4
) -> Firms:
    """Create the firms

    It uses firm_table from rescaleNbFirms

    Parameters
    ----------
    capital_to_value_added_ratio
    firm_table: pandas.DataFrame
        firm_table from rescaleNbFirms
    keep_top_n_firms: None (default) or integer
        (optional) can be specified if we want to keep only the first n firms, for testing purposes
    inventory_restoration_time: float
        Determines the speed at which firms try to reach their inventory duration target
    utilization_rate: float
        Set the utilization rate, which determines the production capacity at the input-output equilibrium.

    Returns
    -------
    list of Firms
    """

    if isinstance(keep_top_n_firms, int):
        firm_table = firm_table.iloc[:keep_top_n_firms, :]

    logging.debug('Creating firms')
    ids = firm_table['id'].tolist()
    firm_table = firm_table.set_index('id')
    if "main_sector" not in firm_table.columns:  # case of MRIO
        firm_table['main_sector'] = firm_table['sector']
    # print(firm_table.head())
    # print(firm_table.iloc[0])
    firms = Firms([
        Firm(i,
             sector=firm_table.loc[i, "sector"],
             sector_type=firm_table.loc[i, "sector_type"],
             main_sector=firm_table.loc[i, "main_sector"],
             od_point=firm_table.loc[i, "od_point"],
             importance=firm_table.loc[i, 'importance'],
             name=firm_table.loc[i, 'name'],
             # geometry=firm_table.loc[i, 'geometry'],
             long=float(firm_table.loc[i, 'long']),
             lat=float(firm_table.loc[i, 'lat']),
             utilization_rate=utilization_rate,
             inventory_restoration_time=inventory_restoration_time,
             capital_to_value_added_ratio=capital_to_value_added_ratio
             )
        for i in ids
    ])
    # We add a bit of noise to the long and lat coordinates
    # It allows to visually disentangle firms located at the same od-point when plotting the map.
    for firm in firms.values():
        firm.add_noise_to_geometry()

    return firms


def define_firms_from_local_economic_data(filepath_region_economic_data: Path,
                                          sectors_to_include: list, transport_nodes: geopandas.GeoDataFrame,
                                          filepath_sector_table: Path, min_nb_firms_per_sector: int):
    """Define firms based on the region_economic_data.
    The output is a dataframe, 1 row = 1 firm.
    The instances Firms are created in the createFirm function.

    Steps:
    1. Load the region_economic_data
    2. It adds a row only when the sector in one region is higher than the sector_cutoffs
    3. It identifies the node of the road network that is the closest to the region point
    4. It combines the firms of the same sector that are in the same road node (case of 2 regions close
    to the same road node)
    5. It calculates the "importance" of each firm = their size relative to the sector size

    Parameters
    ----------
    min_nb_firms_per_sector
    filepath_region_economic_data: string
        Path to the district_data table
    sectors_to_include: list or 'all'
        if 'all', include all sectors, otherwise define the list of sector to include
    transport_nodes: geopandas.GeoDataFrame
        transport nodes resulting from createTransportNetwork
    filepath_sector_table: string
        Path to the sector table
    """

    # A. Create firm table
    # A.1. load files
    region_eco_data = gpd.read_file(filepath_region_economic_data)
    sector_table = pd.read_csv(filepath_sector_table)

    # A.2. for each sector, select region where supply_data is over threshold
    # and populate firm table
    firm_table_per_region = pd.DataFrame()
    for sector, row in sector_table.set_index("sector").iterrows():
        if (sectors_to_include == "all") or (sector in sectors_to_include):
            # check that the supply metric is in the data
            if row["supply_data"] not in region_eco_data.columns:
                logging.warning(f"{row['supply_data']} for sector {sector} is missing from the economic data. "
                                f"We will create by default firms in the {min_nb_firms_per_sector} "
                                f"most populated regions")
                where_create_firm = region_eco_data["population"].nlargest(min_nb_firms_per_sector).index
                # populate firm table
                new_firm_table = pd.DataFrame({
                    "sector": sector,
                    "region": region_eco_data.loc[where_create_firm, "region"].tolist(),
                    "population": region_eco_data.loc[where_create_firm, "population"].tolist(),
                    "absolute_size": region_eco_data.loc[where_create_firm, "population"].tolist()
                })
            else:
                # create one firm where economic metric is over threshold
                where_create_firm = region_eco_data[row["supply_data"]] > row["cutoff"]
                # if it results in less than 5 firms, we go below the cutoff to get at least 5 firms,
                # only if there are enough regions with positive supply_data
                if where_create_firm.sum() < min_nb_firms_per_sector:
                    cond_positive_supply_data = region_eco_data[row["supply_data"]] > 0
                    where_create_firm = region_eco_data.loc[cond_positive_supply_data, row["supply_data"]].nlargest(
                        min_nb_firms_per_sector).index
                # populate firm table
                new_firm_table = pd.DataFrame({
                    "sector": sector,
                    "region": region_eco_data.loc[where_create_firm, "region"].tolist(),
                    "population": region_eco_data.loc[where_create_firm, "population"].tolist(),
                    "absolute_size": region_eco_data.loc[where_create_firm, row["supply_data"]]
                })

            new_firm_table['relative_size'] = new_firm_table['absolute_size'] / new_firm_table['absolute_size'].sum()
            firm_table_per_region = pd.concat([firm_table_per_region, new_firm_table], axis=0)

    # B. Assign firms to the closest road nodes
    # B.1. Create a dictionary that link a region to id of the closest road node
    # Create dic that links regions to points
    selected_regions = list(firm_table_per_region['region'].unique())
    logging.info('Select ' + str(firm_table_per_region.shape[0]) +
                 " in " + str(len(selected_regions)) + ' regions')
    cond = region_eco_data['region'].isin(selected_regions)
    logging.info('Assigning firms to od-points')
    dic_selected_region_to_points = region_eco_data[cond].set_index('region')['geometry'].to_dict()
    # Select road node points
    road_nodes = transport_nodes[transport_nodes['type'] == "roads"]
    # Create dic
    dic_region_to_road_node_id = {
        region: road_nodes.loc[get_index_closest_point(point, road_nodes), 'id']
        for region, point in dic_selected_region_to_points.items()
    }

    # B.2. Map firm to the closest road node
    firm_table_per_region['od_point'] = firm_table_per_region['region'].map(dic_region_to_road_node_id)

    # C. Combine firms that are in the same od-point and in the same sector
    # group by od-point and sector
    firm_table_per_od_point = firm_table_per_region \
        .groupby(['region', 'od_point', 'sector'], as_index=False) \
        .sum()

    # D. Add information required by the createFirms function
    # add sector type
    sector_to_sector_type = sector_table.set_index('sector')['type']
    firm_table_per_od_point['sector_type'] = firm_table_per_od_point['sector'].map(sector_to_sector_type)
    # add long lat
    od_point_table = road_nodes[road_nodes['id'].isin(firm_table_per_od_point['od_point'])].copy()
    od_point_table['long'] = od_point_table.geometry.x
    od_point_table['lat'] = od_point_table.geometry.y
    road_node_id_to_longlat = od_point_table.set_index('id')[['long', 'lat']]
    firm_table_per_od_point['long'] = firm_table_per_od_point['od_point'].map(road_node_id_to_longlat['long'])
    firm_table_per_od_point['lat'] = firm_table_per_od_point['od_point'].map(road_node_id_to_longlat['lat'])
    # add id
    firm_table_per_od_point['id'] = list(range(firm_table_per_od_point.shape[0]))
    # add name, not really useful
    firm_table_per_od_point['name'] = firm_table_per_od_point['od_point'].astype(str) + '-' + firm_table_per_od_point[
        'sector']
    # add importance
    firm_table_per_od_point['importance'] = firm_table_per_od_point['relative_size']

    # # E. Add final demand per firm
    # # evaluate share of population represented
    # cond = region_eco_data['region'].isin(selected_regions)
    # represented_pop = region_eco_data.loc[cond, 'population'].sum()
    # total_population = region_eco_data['population'].sum()
    # # evaluate final demand
    # rel_pop = firm_table['population'] / total_population
    # tot_demand_of_sector = firm_table['sector'].map(sector_table.set_index('sector')['final_demand'])
    # firm_table['final_demand'] = rel_pop * tot_demand_of_sector
    # # print info
    # logging.info("{:.0f}%".format(represented_pop / total_population * 100)+
    #     " of population represented")
    # logging.info("{:.0f}%".format(firm_table['final_demand'].sum() / sector_table['final_demand'].sum() * 100)+
    #     " of final demand is captured")
    # logging.info("{:.0f}%".format(firm_table['final_demand'].sum() / \
    #     sector_table.set_index('sector').loc[sectors_to_include, 'final_demand'].sum() * 100)+
    #     " of final demand of selected sector is captured")

    # F. Log information
    logging.info('Create ' + str(firm_table_per_od_point.shape[0]) + " firms in " +
                 str(firm_table_per_od_point['od_point'].nunique()) + ' od points')
    for sector, row in sector_table.set_index("sector").iterrows():
        if (sectors_to_include == "all") or (sector in sectors_to_include):
            if row["supply_data"] in region_eco_data.columns:
                cond = firm_table_per_od_point['sector'] == sector
                logging.info(f"Sector {sector}: create {cond.sum()} firms that covers " +
                             "{:.0f}%".format(firm_table_per_od_point.loc[cond, 'absolute_size'].sum()
                                              / region_eco_data[row['supply_data']].sum() * 100) +
                             f" of total {row['supply_data']}")
            else:
                cond = firm_table_per_od_point['sector'] == sector
                logging.info(f"Sector {sector}: since {row['supply_data']} is not in the data, "
                             f"create {cond.sum()} firms that covers " +
                             "{:.0f}%".format(firm_table_per_od_point.loc[cond, 'population'].sum()
                                              / region_eco_data["population"].sum() * 100) +
                             f" of population")

    return firm_table_per_od_point, firm_table_per_region


def define_firms_from_mrio_data(
        filepath_country_sector_table: str,
        filepath_region_table: str,
        transport_nodes: pd.DataFrame):
    # Adrien's function for Global

    # Load firm table
    firm_table = pd.read_csv(filepath_country_sector_table)

    # Duplicate the lines by concatenating the DataFrame with itself
    firm_table = pd.concat([firm_table] * 2, ignore_index=True)
    # Assign firms to closest road nodes
    selected_regions = list(firm_table['country_ISO'].unique())
    logging.info('Select ' + str(firm_table.shape[0]) +
                 " firms in " + str(len(selected_regions)) + ' regions')
    location_table = gpd.read_file(filepath_region_table)
    cond_selected_regions = location_table['country_ISO'].isin(selected_regions)
    dic_region_to_points = location_table[cond_selected_regions].set_index('country_ISO')['geometry'].to_dict()
    road_nodes = transport_nodes[transport_nodes['type'] == "roads"]
    dic_region_to_road_node_id = {
        region: road_nodes.loc[get_index_closest_point(point, road_nodes), 'id']
        for region, point in dic_region_to_points.items()
    }
    firm_table['od_point'] = firm_table['country_ISO'].map(dic_region_to_road_node_id)

    # Information required by the createFirms function
    # add long lat
    od_point_table = road_nodes[road_nodes['id'].isin(firm_table['od_point'])].copy()
    od_point_table['long'] = od_point_table.geometry.x
    od_point_table['lat'] = od_point_table.geometry.y
    road_node_id_to_longlat = od_point_table.set_index('id')[['long', 'lat']]
    firm_table['long'] = firm_table['od_point'].map(road_node_id_to_longlat['long'])
    firm_table['lat'] = firm_table['od_point'].map(road_node_id_to_longlat['lat'])
    # add importance
    firm_table['importance'] = 10

    # add id (where is it created otherwise?)
    firm_table = firm_table.reset_index().rename(columns={"index": "id"})

    return firm_table


def check_successful_extraction(firm_table: pd.DataFrame, attribute: str):
    if firm_table[attribute].isnull().sum() > 0:
        logging.warning(f"Unsuccessful extraction of {attribute} for "
                        f"{firm_table.loc[firm_table[attribute].isnull(), 'name']}")


def define_firms_from_mrio(
        filepath_mrio: Path,
        filepath_sector_table: Path,
        filepath_region_table: Path,
        transport_nodes: gpd.GeoDataFrame,
        io_cutoff: float) -> pd.DataFrame:
    # Load mrio
    # mrio = pd.read_csv(filepath_mrio, index_col=0)
    mrio = Mrio.load_mrio_from_filepath(filepath_mrio)
    # Extract region_sectors
    # region_sectors = list(set(mrio.index) | set(mrio.columns))
    # region_sectors = [region_sector for region_sector in region_sectors if
    #                   len(region_sector) == 8]  # format 1011-FRE... :TODO a bit specific to Ecuador, change
    firm_table = pd.DataFrame({
        'tuple': mrio.region_sectors,
        'region': [tup[0] for tup in mrio.region_sectors],
        'main_sector': [tup[1] for tup in mrio.region_sectors],
        'sector': mrio.region_sector_names,
        'name': mrio.region_sector_names
    })
    region_sectors_internal_flows = mrio.get_region_sectors_with_internal_flows(io_cutoff)
    duplicated_firms = pd.DataFrame({
        'tuple': region_sectors_internal_flows,
        'region': [tup[0] for tup in region_sectors_internal_flows],
        'main_sector': [tup[1] for tup in region_sectors_internal_flows],
        'sector': ['_'.join(tup) for tup in region_sectors_internal_flows],
        'name': ['_'.join([tup[0], tup[1], "bis"]) for tup in region_sectors_internal_flows]
    })
    firm_table = pd.concat([firm_table, duplicated_firms])
    # region_sectors = ['_'.join(tup) for tup in mrio.columns]
    # # For region-sector with internal flows, need two firms
    # # region_sectors_internal_flows = list(set(mrio.index) & set(mrio.columns))
    # region_sectors_internal_flows = [tup for tup in mrio.columns if mrio.loc[tup, tup] > 0]
    # region_sectors = region_sectors + [region_sector + '-bis' for region_sector in region_sectors_internal_flows]
    # # Create firm_table
    # firm_table = pd.DataFrame({"name": region_sectors})
    # # firm_table['region'] = firm_table['name'].str.extract(r'([0-9]*)-[A-Z0-9]{3}')
    # firm_table['region'] = firm_table['name'].str.extract(r'([A-Z]{3})')
    # check_successful_extraction(firm_table, "region")
    # # firm_table['main_sector'] = firm_table['name'].str.extract(r'[0-9]*-([A-Z]{2}[A-Z0-9]{1})')
    # firm_table['main_sector'] = firm_table['name'].str[4:]
    # check_successful_extraction(firm_table, "main_sector")
    # firm_table['sector'] = firm_table['name'].str.replace('-bis', '')
    # check_successful_extraction(firm_table, "sector")
    logging.info(f"Select {firm_table.shape[0]} firms in {firm_table['region'].nunique()} regions")

    # Assign firms to the nearest road node
    firm_table['od_point'] = get_closest_road_nodes(firm_table['region'], transport_nodes, filepath_region_table)

    # Add long lat
    long_lat = get_long_lat(firm_table['od_point'], transport_nodes)
    firm_table['long'] = long_lat['long']
    firm_table['lat'] = long_lat['lat']

    # Add importance
    # row_intermediary = [row for row in mrio.index if len(row) == 8]
    tot_outputs_per_region_sector = mrio.sum(axis=1)
    firm_table['importance'] = firm_table['tuple'].map(tot_outputs_per_region_sector)
    firm_with_two_firms_same_region_sector = firm_table['tuple'].isin(region_sectors_internal_flows)
    firm_table.loc[firm_with_two_firms_same_region_sector, "importance"] = \
        firm_table.loc[firm_with_two_firms_same_region_sector, "importance"] / 2
    check_successful_extraction(firm_table, "importance")

    # Add sector type
    sector_table = pd.read_csv(filepath_sector_table)
    firm_table['sector_type'] = firm_table['sector'].map(sector_table.set_index('sector')['type'])
    check_successful_extraction(firm_table, "sector_type")

    # Add id (where is it created otherwise?)
    firm_table['id'] = range(firm_table.shape[0])

    return firm_table


def define_firms_from_network_data(
        filepath_firm_table: str,
        filepath_location_table: str,
        sectors_to_include: list,
        transport_nodes: pd.DataFrame,
        filepath_sector_table: str
):
    """Define firms based on the firm_table
    The output is a dataframe, 1 row = 1 firm.
    The instances Firms are created in the createFirm function.
    """
    # Load firm table
    firm_table = pd.read_csv(filepath_firm_table, dtype={'region': str})

    # Filter out some sectors
    if sectors_to_include != "all":
        firm_table = firm_table[firm_table['sector'].isin(sectors_to_include)]

    # Assign firms to closest road nodes
    selected_regions = list(firm_table['region'].unique())
    logging.info('Select ' + str(firm_table.shape[0]) +
                 " firms in " + str(len(selected_regions)) + ' regions')
    location_table = gpd.read_file(filepath_location_table)
    cond_selected_regions = location_table['region'].isin(selected_regions)
    dic_region_to_points = location_table[cond_selected_regions].set_index('region')['geometry'].to_dict()
    road_nodes = transport_nodes[transport_nodes['type'] == "roads"]
    dic_region_to_road_node_id = {
        region: road_nodes.loc[get_index_closest_point(point, road_nodes), 'id']
        for region, point in dic_region_to_points.items()
    }
    firm_table['od_point'] = firm_table['region'].map(dic_region_to_road_node_id)

    # Information required by the createFirms function
    # add sector type
    sector_table = pd.read_csv(filepath_sector_table)
    sector_to_sector_type = sector_table.set_index('sector')['type']
    firm_table['sector_type'] = firm_table['sector'].map(sector_to_sector_type)
    # add long lat
    od_point_table = road_nodes[road_nodes['id'].isin(firm_table['od_point'])].copy()
    od_point_table['long'] = od_point_table.geometry.x
    od_point_table['lat'] = od_point_table.geometry.y
    road_node_id_to_longlat = od_point_table.set_index('id')[['long', 'lat']]
    firm_table['long'] = firm_table['od_point'].map(road_node_id_to_longlat['long'])
    firm_table['lat'] = firm_table['od_point'].map(road_node_id_to_longlat['lat'])
    # add importance
    firm_table['importance'] = 10

    return firm_table


def load_technical_coefficients(
        firms: Firms,
        filepath_tech_coef: str,
        io_cutoff: float = 0.1,
        import_sector_name: str | None = "IMP"
):
    """Load the input mix of the firms' Leontief function

    Parameters
    ----------
    firms : pandas.DataFrame
        the list of Firms generated from the createFirms function

    filepath_tech_coef : string
        Filepath to the matrix of technical coefficients

    io_cutoff : float
        Filters out technical coefficient below this cutoff value. Default to 0.1.

    import_sector_name : None or string
        Give the name of the import sector. If None, then the import technical coefficient is discarded. Default to None

    Returns
    -------
    list of Firms
    """

    # Load technical coefficient matrix from data
    tech_coef_matrix = pd.read_csv(filepath_tech_coef, index_col=0)
    tech_coef_matrix = tech_coef_matrix.mask(tech_coef_matrix <= io_cutoff, 0)

    # We select only the technical coefficient between sectors that are actually represented in the economy
    # Note that, when filtering out small sector-district combination, some sector may not be present.
    sector_present = list(set([firm.sector for firm in firms.values()]))
    if import_sector_name:
        tech_coef_matrix = tech_coef_matrix.loc[sector_present + [import_sector_name], sector_present]
    else:
        tech_coef_matrix = tech_coef_matrix.loc[sector_present, sector_present]

    # Check whether all sectors have input
    cond_sector_no_inputs = tech_coef_matrix.sum() == 0
    if cond_sector_no_inputs.any():
        warnings.warn(
            'Some sectors have no inputs: ' + str(cond_sector_no_inputs[cond_sector_no_inputs].index.to_list())
            + " Check this sector or reduce the io_coef cutoff")

    # Load input mix
    for firm in firms.values():
        firm.input_mix = tech_coef_matrix.loc[tech_coef_matrix.loc[:, firm.sector] != 0, firm.sector].to_dict()

    logging.info('Technical coefficient loaded. io_cutoff: ' + str(io_cutoff))


def load_mrio_tech_coefs(
        firms: Firms,
        filepath_mrio: Path,
        io_cutoff: float
):
    # Load mrio
    mrio = Mrio.load_mrio_from_filepath(filepath_mrio)
    # mrio = pd.read_csv(filepath_mrio, header=[0, 1], index_col=[0, 1])  # TODO: class mrio
    # mrio = pd.read_csv(filepath_mrio, index_col=0)

    # Load technical coefficient matrix from data
    # region_sectors = [tup for tup in mrio.index if tup[1] != "Imports"]
    # tot_outputs = mrio.loc[region_sectors].sum(axis=1)
    # matrix_output = pd.concat([tot_outputs] * len(mrio.index), axis=1).transpose()
    # matrix_output.index = mrio.index
    # tech_coef_matrix = mrio[region_sectors] / matrix_output
    tech_coef_dict = mrio.get_tech_coef_dict(threshold=io_cutoff)

    # Load into firms
    for firm in firms.values():
        if firm.sector in tech_coef_dict.keys():
            firm.input_mix = tech_coef_dict[firm.sector]
        else:
            firm.input_mix = {}

    logging.info('Technical coefficient loaded.')


def calibrate_input_mix(
        firms: Firms,
        firm_table: pd.DataFrame,
        sector_table: pd.DataFrame,
        filepath_transaction_table: str
):
    transaction_table = pd.read_csv(filepath_transaction_table)

    domestic_b2b_sales_per_firm = transaction_table.groupby('supplier_id')['transaction'].sum()
    firm_table['domestic_B2B_sales'] = firm_table['id'].map(domestic_b2b_sales_per_firm).fillna(0)
    firm_table['output'] = firm_table['domestic_B2B_sales'] + firm_table['final_demand'] + firm_table['exports']

    # Identify the sector of the products exchanged recorded in the transaction table and whether they are essential
    transaction_table['product_sector'] = transaction_table['supplier_id'].map(firm_table.set_index('id')['sector'])
    transaction_table['is_essential'] = transaction_table['product_sector'].map(
        sector_table.set_index('sector')['essential'])

    # Get input mix from this data
    def get_input_mix(transaction_from_unique_buyer, firm_tab):
        output = firm_tab.set_index('id').loc[transaction_from_unique_buyer.name, 'output']
        print(output)
        cond_essential = transaction_from_unique_buyer['is_essential']
        # for essential inputs, get total input per product type
        input_essential = transaction_from_unique_buyer[cond_essential].groupby('product_sector')[
                              'transaction'].sum() / output
        # for non essential inputs, get total input
        input_nonessential = transaction_from_unique_buyer.loc[~cond_essential, 'transaction'].sum() / output
        # get share how much is essential and evaluate how much can be produce with essential input only (beta)
        share_essential = input_essential.sum() / transaction_from_unique_buyer['transaction'].sum()
        max_output_with_essential_only = share_essential * output
        # shape results
        dic_res = input_essential.to_dict()
        dic_res['non_essential'] = input_nonessential
        dic_res['max_output_with_essential_only'] = max_output_with_essential_only
        return dic_res

    input_mix = transaction_table.groupby('buyer_id').apply(get_input_mix, firm_table)

    # Load input mix into Firms
    for firm in firms.values():
        firm.input_mix = input_mix[firm.pid]

    return firms, transaction_table


def load_inventories(firms: Firms, inventory_duration_target: int | str, given_time_unit: str, model_time_unit: str,
                     filepath_inventory_duration_targets: Path, extra_inventory_target: int | None = None,
                     inputs_with_extra_inventories: None | list = None,
                     buying_sectors_with_extra_inventories: None | list = None,
                     min_inventory: int = 1):
    """Load inventory duration target

    If inventory_duration_target is an integer, it is uniformly applied to all firms.
    If it its "inputed", then we use the targets defined in the file filepath_inventory_duration_targets. In that case,
    targets are sector specific, i.e., it varies according to the type of input and the sector of the buying firm.
    If both cases, we can add extra units of inventories:
    - uniformly, e.g., all firms have more inventories of all inputs,
    - to specific inputs, all firms have extra agricultural inputs,
    - to specific buying firms, e.g., all manufacturing firms have more of all inputs,
    - to a combination of both. e.g., all manufacturing firms have more of agricultural inputs.
    We can also add some noise on the distribution of inventories. Not yet implemented.

    Parameters
    ----------
    model_time_unit
    given_time_unit
    filepath_inventory_duration_targets
    firms : pandas.DataFrame
        the list of Firms generated from the createFirms function

    inventory_duration_target : "inputed" or integer
        Inventory duration target uniformly applied to all firms and all inputs.
        If 'inputed', uses the specific values from the file specified by
        filepath_inventory_duration_targets

    extra_inventory_target : None or integer
        If specified, extra inventory duration target.

    inputs_with_extra_inventories : None or list of sector
        For which inputs do we add inventories.

    buying_sectors_with_extra_inventories : None or list of sector
        For which sector we add inventories.

    min_inventory : int
        Set a minimum inventory level
    """
    time_unit_in_days = {
        "day": 1,
        "week": 7,
        "month": 30,
        "year": 365
    }
    time_adjustment = time_unit_in_days[given_time_unit] / time_unit_in_days[model_time_unit]

    if isinstance(inventory_duration_target, int):
        for firm in firms.values():
            firm.inventory_duration_target = {input_sector: time_adjustment * inventory_duration_target
                                              for input_sector in firm.input_mix.keys()}

    elif inventory_duration_target == 'inputed':
        dic_sector_inventory = pd.read_csv(filepath_inventory_duration_targets) \
                               .set_index(['buying_sector', 'input_sector'])['inventory_duration_target'].to_dict()
        for firm in firms.values():
            firm.inventory_duration_target = {
                input_sector: time_adjustment * dic_sector_inventory[(firm.sector, input_sector)]
                for input_sector in firm.input_mix.keys()
            }

    else:
        raise ValueError("Unknown value entered for 'inventory_duration_target'")

    # if random_mean_sd:
    #     if random_draw:
    #         for firm in firms:
    #             firm.inventory_duration_target = {}
    #             for input_sector in firm.input_mix.keys():
    #                 mean = dic_sector_inventory[(firm.sector, input_sector)]['mean']
    #                 sd = dic_sector_inventory[(firm.sector, input_sector)]['sd']
    #                 mu = math.log(mean/math.sqrt(1+sd**2/mean**2))
    #                 sigma = math.sqrt(math.log(1+sd**2/mean**2))
    #                 safety_day = np.random.log(mu, sigma)
    #                 firm.inventory_duration_target[input_sector] = safety_day

    # Add extra inventories if needed. Not the best programming maybe...
    if isinstance(extra_inventory_target, int):
        if isinstance(inputs_with_extra_inventories, list) and (buying_sectors_with_extra_inventories == 'all'):
            for firm in firms.values():
                firm.inventory_duration_target = {
                    input_sector: firm.inventory_duration_target[input_sector] + extra_inventory_target*time_adjustment
                    if (input_sector in inputs_with_extra_inventories) else firm.inventory_duration_target[input_sector]
                    for input_sector in firm.input_mix.keys()
                }

        elif (inputs_with_extra_inventories == 'all') and isinstance(buying_sectors_with_extra_inventories, list):
            for firm in firms.values():
                firm.inventory_duration_target = {
                    input_sector: firm.inventory_duration_target[input_sector] + extra_inventory_target*time_adjustment
                    if (firm.sector in buying_sectors_with_extra_inventories) else firm.inventory_duration_target[
                        input_sector]
                    for input_sector in firm.input_mix.keys()
                }

        elif isinstance(inputs_with_extra_inventories, list) and isinstance(buying_sectors_with_extra_inventories,
                                                                            list):
            for firm in firms.values():
                firm.inventory_duration_target = {
                    input_sector: firm.inventory_duration_target[input_sector] + extra_inventory_target*time_adjustment
                    if ((input_sector in inputs_with_extra_inventories) and (
                            firm.sector in buying_sectors_with_extra_inventories)) else
                    firm.inventory_duration_target[input_sector]
                    for input_sector in firm.input_mix.keys()
                }

        elif (inputs_with_extra_inventories == 'all') and (buying_sectors_with_extra_inventories == 'all'):
            for firm in firms.values():
                firm.inventory_duration_target = {
                    input_sector: firm.inventory_duration_target[input_sector] + extra_inventory_target*time_adjustment
                    for input_sector in firm.input_mix.keys()
                }

        else:
            raise ValueError("Unknown value given for 'inputs_with_extra_inventories' or "
                             "'buying_sectors_with_extra_inventories'. Should be a list of string or 'all'")

    if min_inventory > 0:
        for firm in firms.values():
            firm.inventory_duration_target = {
                input_sector: max(min_inventory * time_adjustment, inventory)
                for input_sector, inventory in firm.inventory_duration_target.items()
            }

    logging.info('Inventory duration targets loaded')
    if extra_inventory_target:
        logging.info(f"Extra inventory duration: {extra_inventory_target} "
                     f"for inputs {inputs_with_extra_inventories} "
                     f"for buying sectors{buying_sectors_with_extra_inventories}")



import json
import logging
import os
from pathlib import Path
import pandas as pd
import geopandas as gpd

from src.network.sc_network import ScNetwork


class Simulation(object):
    def __init__(self, simulation_type: str):
        if simulation_type not in ["initial_state", "event"]:
            raise ValueError("Simulation type should be initial_state or event")
        self.type = simulation_type
        self.firm_data = []
        self.country_data = []
        self.household_data = []
        self.sc_network_data = []
        self.transport_network_data = []

    def export_agent_data(self, export_folder):
        logging.info(f'Exporting agent data to {export_folder}')
        with open(os.path.join(export_folder, 'firm_data.json'), 'w') as jsonfile:
            json.dump(self.firm_data, jsonfile)
        with open(os.path.join(export_folder, 'country_data.json'), 'w') as jsonfile:
            json.dump(self.country_data, jsonfile)
        with open(os.path.join(export_folder, 'household_data.json'), 'w') as jsonfile:
            json.dump(self.household_data, jsonfile)

    def export_transport_network_data(self, transport_edges: gpd.GeoDataFrame, export_folder: Path):
        logging.info(f'Exporting transport network data to {export_folder}')
        flow_df = pd.DataFrame(self.transport_network_data)
        for time_step in flow_df['time_step'].unique():
            transport_edges_with_flows = pd.merge(
                transport_edges, flow_df[flow_df['time_step'] == time_step],
                how="left", on="id")
            transport_edges_with_flows.to_file(export_folder / f"transport_edges_with_flows_{time_step}.geojson",
                                               driver="GeoJSON", index=False)

    def calculate_and_export_summary_result(self, sc_network: ScNetwork, household_table: pd.DataFrame,
                                            monetary_unit_in_model: str, export_folder: Path):
        if self.type == "initial_state":
            # export io matrix
            logging.info(f'Exporting resulting IO matrix to {export_folder}')
            sc_network.calculate_io_matrix().to_csv(export_folder / "io_table.csv")
            logging.info(f'Exporting edgelist to {export_folder}')
            sc_network.generate_edge_list().to_csv(export_folder / "sc_network_edgelist.csv")

        elif self.type == "event":
            # export loss time series for households
            logging.info(f'Exporting loss time series of households per region sector to {export_folder}')
            household_result_table = pd.DataFrame(self.household_data)
            loss_per_region_sector_time = household_result_table.groupby('household').apply(
                self.summarize_results_one_household).reset_index().drop(columns=['level_1'])
            household_table['id'] = 'hh_' + household_table['id'].astype(str)
            loss_per_region_sector_time['region'] = loss_per_region_sector_time['household'].map(
                household_table.set_index('id')['region'])
            loss_per_region_sector_time = \
                loss_per_region_sector_time.groupby(['region', 'sector', 'time_step'], as_index=False)['loss'].sum()
            loss_per_region_sector_time.to_csv(export_folder / "loss_per_region_sector_time.csv", index=False)
            household_loss = loss_per_region_sector_time['loss'].sum()
            logging.info(f'Exporting loss time series of countries to {export_folder}')
            # export loss time series for countries
            country_result_table = pd.DataFrame(self.country_data)
            country_result_table['loss'] = country_result_table['extra_spending'] \
                                           + country_result_table['consumption_loss']
            country_result_table = country_result_table[['time_step', 'country', 'loss']]
            country_loss = country_result_table['loss'].sum()
            country_result_table.to_csv(export_folder / "loss_per_country.csv", index=False)
            logging.info(f"Cumulated household loss: {household_loss:,.2f} {monetary_unit_in_model}")
            logging.info(f"Cumulated country loss: {country_loss:,.2f} {monetary_unit_in_model}")
            # Export summary
            total_loss = pd.DataFrame({"households": household_loss, "countries": country_loss}, index=[0])
            total_loss.to_csv(export_folder / "loss_summary.csv", index=False)

    @staticmethod
    def summarize_results_one_household(household_result_table_one_household):
        extra_spending_per_sector_table = pd.DataFrame(
            household_result_table_one_household.set_index('time_step')['extra_spending_per_sector'].to_dict()
        ).transpose()
        consumption_loss_per_sector_table = pd.DataFrame(
            household_result_table_one_household.set_index('time_step')['consumption_loss_per_sector'].to_dict()
        ).transpose()
        loss_per_sector = extra_spending_per_sector_table + consumption_loss_per_sector_table
        result = loss_per_sector.stack().reset_index()
        result.columns = ['time_step', 'sector', 'loss']
        return result



import importlib
import logging
import os
from datetime import datetime
from pathlib import Path

import yaml
from dataclasses import dataclass

from src import paths

EPSILON = 1e-6
import_code = "IMP"

@dataclass
class Parameters:
    scope: str
    export_details: dict
    specific_edges_to_monitor: dict
    logging_level: str
    transport_modes: list
    monetary_units_in_model: str
    monetary_units_inputed: str
    firm_data_type: str
    congestion: bool
    propagate_input_price_change: bool
    sectors_to_include: str
    sectors_to_exclude: list | None
    sectors_no_transport_network: list
    cutoff_sector_output: dict
    cutoff_sector_demand: dict
    combine_sector_cutoff: str
    districts_to_include: str | list
    pop_density_cutoff: float
    pop_cutoff: float
    min_nb_firms_per_sector: int
    local_demand_cutoff: float
    countries_to_include: str | list
    district_sector_cutoff: str
    nb_top_district_per_sector: None | int
    explicit_service_firm: bool
    inventory_duration_target: str | int
    extra_inventory_target: None | int
    inputs_with_extra_inventories: str | list
    buying_sectors_with_extra_inventories: str | list
    inventory_restoration_time: float
    utilization_rate: float
    io_cutoff: float
    rationing_mode: str
    nb_suppliers_per_input: float
    weight_localization_firm: float
    weight_localization_household: float
    force_local_retailer: bool
    events: list
    time_resolution: str
    inventory_duration_target_unit: str
    nodeedge_tested_topn: None | int
    nodeedge_tested_skipn: None | int
    model_IO: bool
    duration_dic: dict
    extra_roads: bool
    epsilon_stop_condition: float
    route_optimization_weight: str
    cost_repercussion_mode: str
    price_increase_threshold: float
    capacity_constraint: bool
    transport_cost_noise_level: float
    firm_sampling_mode: str
    filepaths: dict
    export_files: bool
    simulation_type: str
    adaptive_inventories: bool
    adaptive_supplier_weight: bool
    transport_cost_data: dict
    capital_to_value_added_ratio: float
    export_folder: Path | str = ""

    @classmethod
    def load_default_parameters(cls, parameter_folder: Path):
        with open(parameter_folder / "default.yaml", 'r') as f:
            default_parameters = yaml.safe_load(f)
        return cls(**default_parameters)

    @classmethod
    def load_parameters(cls, parameter_folder: Path, scope: str):
        # Load default and user_defined parameters
        with open(parameter_folder / "default.yaml", 'r') as f:
            parameters = yaml.safe_load(f)
        user_defined_parameter_filepath = parameter_folder / f"user_defined_{scope}.yaml"
        if os.path.exists(user_defined_parameter_filepath):
            logging.info(f'User defined parameter file found for {scope}')
            with open(parameter_folder / f"user_defined_{scope}.yaml", 'r') as f:
                overriding_parameters = yaml.safe_load(f)
            # Merge both
            for key, val in parameters.items():
                if key in overriding_parameters:
                    if isinstance(val, dict):
                        cls.merge_dict_with_priority(parameters[key], overriding_parameters[key])
                    else:
                        parameters[key] = overriding_parameters[key]
        else:
            logging.info(f'No user defined parameter file found named user_defined_{scope}.yaml, '
                         f'using default parameters')
        # Load scope
        parameters['scope'] = scope
        # Create parameters
        parameters = cls(**parameters)
        # Adjust filepath
        parameters.build_full_filepath()
        # Create export folder

        # Cast datatype
        parameters.epsilon_stop_condition = float(parameters.epsilon_stop_condition)
        parameters.duration_dic = {int(key): val for key, val in parameters.duration_dic.items()}

        return parameters

    @staticmethod
    def merge_dict_with_priority(default_dict: dict, overriding_dict: dict):
        for key, val in default_dict.items():
            if key in overriding_dict:
                default_dict[key] = overriding_dict[key]

    def get_full_filepath(self, filepath):
        return paths.INPUT_FOLDER / self.scope / filepath

    def build_full_filepath(self):
        for key, val in self.filepaths.items():
            if val == "None":
                self.filepaths[key] = None
            else:
                self.filepaths[key] = self.get_full_filepath(val)
        if self.events:
            for event in self.events:
                for key, item in event.items():
                    if "filepath" in key:
                        event[key] = self.get_full_filepath(item)

    def export(self):
        with open(self.export_folder / 'parameters.yaml', 'w') as file:
            yaml.dump(self, file)

    def create_export_folder(self):
        if not os.path.isdir(paths.OUTPUT_FOLDER / self.scope):
            os.mkdir(paths.OUTPUT_FOLDER / self.scope)
        self.export_folder = paths.OUTPUT_FOLDER / self.scope / datetime.now().strftime('%Y%m%d_%H%M%S')
        os.mkdir(self.export_folder)

    def adjust_logging_behavior(self):
        if self.logging_level == "info":
            logging_level = logging.INFO
        else:
            logging_level = logging.DEBUG

        if self.export_files:
            importlib.reload(logging)
            logging.basicConfig(
                filename=self.export_folder / 'exp.log',
                level=logging_level,
                format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
            )
            logging.getLogger().addHandler(logging.StreamHandler())
        else:
            importlib.reload(logging)
            logging.basicConfig(
                level=logging_level,
                format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
            )



from typing import TYPE_CHECKING

import logging
from collections import UserList
from pathlib import Path

import geopandas
import pandas
import pandas as pd

from src.parameters import EPSILON, import_code

if TYPE_CHECKING:
    from src.agents.firm import Firms
    from src.network.transport_network import TransportNetwork
    from src.model.model import Model


class ReconstructionMarket:
    def __init__(self, reconstruction_target_time: int, capital_input_mix: dict):
        self.reconstruction_target_time = reconstruction_target_time
        self.capital_input_mix = capital_input_mix
        self.aggregate_demand = 0
        self.aggregate_demand_per_sector = {}
        self.demand_to_firm_per_sector = {}

    def send_orders(self, firms: "Firms"):
        for sector, demand_to_firm_this_sector in self.demand_to_firm_per_sector.items():
            for pid, reconstruction_demand in demand_to_firm_this_sector.items():
                firms[pid].reconstruction_demand = reconstruction_demand
                firms[pid].add_reconstruction_order_to_order_book()

    def distribute_new_capital(self, firms: "Firms"):
        # Retrieve production
        print("total capital demanded:", self.aggregate_demand)
        amount_produced_per_sector = {}
        for sector in self.capital_input_mix.keys():
            if sector == import_code:
                amount_produced_per_sector[sector] = self.aggregate_demand_per_sector[sector]
            else:
                amount_produced_per_sector[sector] = sum([firm.reconstruction_produced
                                                          for firm in firms.filter_by_sector(sector).values()])
        print("amount_produced_per_sector:", amount_produced_per_sector)
        # Produce (we suppose that what is not used disappear, no stock of unfinished capital)
        new_capital_produced = min([amount_produced_per_sector[sector] / weight
                                    for sector, weight in self.capital_input_mix.items()])
        print("new_capital_produced:", new_capital_produced)
        # Send new capital to firm
        for firm in firms.values():
            firm.capital_destroyed -= (firm.capital_demanded / self.aggregate_demand) * new_capital_produced
        print("total capital destroyed:", firms.sum("capital_destroyed"))
        print("average production capacity:", firms.mean("current_production_capacity"))

    def evaluate_demand_to_firm(self, firms: "Firms"):
        # Retrieve the demand of each firm, and translate it into a demand for certain inputs (sectors)
        for firm in firms.values():
            firm.capital_demanded = firm.capital_destroyed / self.reconstruction_target_time
        self.aggregate_demand = sum([firm.capital_demanded for firm in firms.values()])
        self.aggregate_demand_per_sector = {sector: weight * self.aggregate_demand
                                            for sector, weight in self.capital_input_mix.items()}

        # Get potential supply per sector and evaluate whether demand needs to be rationed
        rationing_per_sector = {}
        total_supply_per_sector = {}
        potential_supply_per_firm_per_sector = {}
        for sector in self.capital_input_mix.keys():
            if sector == import_code:  # No constraints for imported products
                total_supply_per_sector[sector] = self.aggregate_demand_per_sector[sector]
                rationing_per_sector[sector] = 1
            else:
                potential_supply_per_firm_per_sector[sector] = {pid: firm.get_spare_production_potential()
                                                                for pid, firm in firms.items()
                                                                if firm.sector == sector}
                total_supply_per_sector[sector] = sum(potential_supply_per_firm_per_sector[sector].values())
                rationing_per_sector[sector] = min(1, total_supply_per_sector[sector]
                                                   / self.aggregate_demand_per_sector[sector])
        rationing = min(list(rationing_per_sector.values()))
        if rationing > 1 - EPSILON:
            logging.info("Reconstruction market: There is no rationing")
        else:
            logging.info(f"Reconstruction market: Due to limited capacity, "
                         f"supply for reconstruction is {rationing:.2%} of demand")

        # Evaluate actual demand per firm
        for sector in self.capital_input_mix.keys():
            if sector == import_code:
                self.demand_to_firm_per_sector[sector] = {}
            else:
                if rationing < EPSILON:
                    self.demand_to_firm_per_sector[sector] = {}
                else:
                    adjusted_demand_this_sector = self.aggregate_demand_per_sector[sector] * rationing
                    firm_weight = {pid: potential_supply / total_supply_per_sector[sector]
                                   for pid, potential_supply in potential_supply_per_firm_per_sector[sector].items()}
                    self.demand_to_firm_per_sector[sector] = {pid: weight * adjusted_demand_this_sector
                                                              for pid, weight in firm_weight.items()}


class Recovery:
    def __init__(self, duration: int, shape: str):
        self.duration = duration
        self.shape = shape


class TransportDisruption(dict):
    def __init__(self, description: dict, recovery: Recovery | None = None, start_time: int = 0):
        self.start_time = start_time
        self.recovery = recovery
        if description is not None:
            for key, value in description.items():
                self.__setitem__(key, value)
        super().__init__()

    def __setitem__(self, key, value):
        if not isinstance(key, int):
            raise KeyError("Key must be a int: the id of the transport edge to be disrupted")
        if not isinstance(value, float):
            raise ValueError("Value must be a float: the fraction of lost capacity")
        super().__setitem__(key, value)

    def __repr__(self):
        return f"EventDict(start_time={self.start_time}, data={super().__repr__()})"

    def log_info(self):
        if self.recovery:
            logging.info(f"{len(self)} transport edges are disrupted at {self.start_time} time steps. "
                         f"They recover after {self.recovery.duration} time steps.")
        else:
            logging.info(f"{len(self)} transport edges are disrupted at {self.start_time} time steps. "
                         f"There is no recovery.")

    @classmethod
    def from_edge_attributes(cls, edges: geopandas.GeoDataFrame, attribute: str, values: list):
        # we do a special case for the disruption attribute
        # for which we check if the attribute contains one of the value
        if attribute == "disruption":
            condition = [edges[attribute].str.contains(value) for value in values]
            condition = pandas.concat(condition, axis=1)
            condition = condition.any(axis=1)
        else:
            condition = edges[attribute].isin(values)
        item_ids = edges.sort_values('id').loc[condition, 'id'].tolist()
        description = pd.Series(1.0, index=item_ids).to_dict()

        return cls(
            description=description,
        )

    def implement(self, transport_network: "TransportNetwork"):
        for edge in transport_network.edges:
            edge_id = transport_network[edge[0]][edge[1]]['id']
            if edge_id in self.keys():
                transport_network.disrupt_one_edge(edge, self.recovery.duration, self[edge_id])


class CapitalDestruction(dict):
    def __init__(self, description: dict, recovery: Recovery | None = None, start_time: int = 1,
                 reconstruction_market: bool = False):
        self.start_time = start_time
        self.recovery = recovery
        self.reconstruction_market = reconstruction_market
        if description is not None:
            for key, value in description.items():
                self.__setitem__(key, value)
        super().__init__()

    def __setitem__(self, key, value):
        if not isinstance(key, int):
            raise KeyError("Key must be an int: the id of the firm")
        if not isinstance(value, float):
            raise ValueError("Value must be a float: the amount of destroyed capital")
        super().__setitem__(key, value)

    def __repr__(self):
        return f"EventDict(start_time={self.start_time}, data={super().__repr__()})"

    def log_info(self):
        logging.info(f"{len(self)} firms incur capital destruction at {self.start_time} time steps. "
                     f"There is {'a' if self.reconstruction_market else 'no'} reconstruction market.")
        if self.recovery:
            logging.info(f"There is exogenous recovery after {self.recovery.duration} time steps.")
        else:
            logging.info("There is no exogenous recovery.")

    @classmethod
    def from_region_sector_file(cls, filepath: Path, firm_table: pd.DataFrame, firm_list: "Firms",
                                input_units: str, target_units: str):
        df = pd.read_csv(filepath, dtype={'region': str, 'sector': str, 'destroyed_capital': float})
        units = {"USD": 1, "kUSD": 1e3, "mUSD": 1e6}
        df['destroyed_capital'] = df['destroyed_capital'] * units[input_units] / units[target_units]
        result = {}
        dic_destroyed_capital_per_region_sector = df.set_index(['region', 'sector'])['destroyed_capital'].to_dict()
        for region_sector, destroyed_capital in dic_destroyed_capital_per_region_sector.items():
            region, sector = region_sector
            firms = firm_table.loc[(firm_table['region'] == region) & (firm_table['sector'] == sector), "id"].to_list()
            if len(firms) == 0:
                logging.warning(f"In {region_sector}, destroyed capital is {destroyed_capital} {input_units} "
                                f"but there are no firm modeled")
            else:
                total_capital = sum([firm_list[firm_id].capital_initial for firm_id in firms])
                for firm_id in firms:
                    weight = firm_list[firm_id].capital_initial / total_capital
                    result[firm_id] = weight * destroyed_capital
        df = pd.merge(df, firm_table[['region', 'sector', 'id']], how='left', on=['region', 'sector'])
        total_destroyed_capital_in_data = df['destroyed_capital'].sum()
        total_destroyed_capital_in_model = sum(result.values())
        logging.info(f"Destroyed capital in data: {total_destroyed_capital_in_data}, "
                     f"Destroyed capital in model: {total_destroyed_capital_in_model}")
        return cls(
            description=result,
            recovery=None
        )

    def implement(self, firm_list: "Firms", model: "Model"):
        for firm_id, destroyed_capital in self.items():
            firm_list[firm_id].incur_capital_destruction(destroyed_capital)
        if self.reconstruction_market:
            model.reconstruction_market = ReconstructionMarket(reconstruction_target_time=30,
                                                               capital_input_mix={"CON": 0.7, "MAN": 0.2, "IMP": 0.1})


class DisruptionList(UserList):
    def __init__(self, disruption_list: list):
        super().__init__(disruption for disruption in disruption_list
                         if isinstance(disruption, CapitalDestruction) or isinstance(disruption, TransportDisruption))
        if len(disruption_list) > 0:
            self.start_time = min([disruption.start_time for disruption in disruption_list])
            self.end_time = 0
            # self.end_time = max([disruption.start_time + disruption.duration for disruption in disruption_list])
            # self.transport_nodes = [
            #     disruption.item_id
            #     for disruption in disruption_list
            #     if disruption.item_type == "transport_node"
            # ]
            # self.transport_edges = [
            #     disruption.item_id
            #     for disruption in disruption_list
            #     if disruption.item_type == "transport_edge"
            # ]
            # self.firms = [
            #     disruption.item_id
            #     for disruption in disruption_list
            #     if disruption.item_type == "firm"
            # ]
        else:
            self.start_time = 0
            self.end_time = 0

    @classmethod
    def from_events_parameter(
            cls,
            events: list,
            model_unit: str,
            edges: geopandas.GeoDataFrame,
            firm_table: pandas.DataFrame,
            firm_list: "Firms"
    ):
        event_list = []
        for event in events:
            if event['type'] == "capital_destruction":
                if event['description_type'] == "region_sector_file":
                    disruption_object = CapitalDestruction.from_region_sector_file(event['region_sector_filepath'],
                                                                                   firm_table,
                                                                                   firm_list,
                                                                                   input_units=event['unit'],
                                                                                   target_units=model_unit)
                    disruption_object.start_time = event["start_time"]
                    if "reconstruction_market" in event.keys():
                        disruption_object.reconstruction_market = event["reconstruction_market"]
                    event_list += [disruption_object]
            if event['type'] == "transport_disruption":
                if event['description_type'] == "edge_attributes":
                    disruption_object = TransportDisruption.from_edge_attributes(
                        edges=edges,
                        attribute=event['attribute'],
                        values=event['values']
                    )
                    disruption_object.start_time = event["start_time"]
                    disruption_object.recovery = Recovery(duration=event['duration'], shape="threshold")
                    event_list += [disruption_object]
        return cls(event_list)

    def log_info(self):
        logging.info(f'There are {len(self)} disruptions')
        for disruption in self:
            disruption.log_info()

    # def filter_type(self, selected_item_type):
    #     return DisruptionList([disruption for disruption in self if disruption.item_type == selected_item_type])

    def filter_start_time(self, selected_start_time):
        return DisruptionList([disruption for disruption in self if disruption.start_time == selected_start_time])

    def get_item_id_duration_reduction_dict(self) -> dict:
        return {
            disruption.item_id: {
                "duration": disruption.duration,
                "reduction": disruption.reduction
            }
            for disruption in self
        }

    def get_id_list(self) -> list:
        return [disruption.item_id for disruption in self]



from typing import TYPE_CHECKING

import numpy as np

from src.model.basic_functions import calculate_distance_between_agents, rescale_values, generate_weights, \
    add_or_increment_dict_key

import logging

from src.agents.agent import Agent, Agents
from src.network.commercial_link import CommercialLink
from src.network.mrio import import_label

if TYPE_CHECKING:
    from src.network.sc_network import ScNetwork
    from src.agents.firm import Firms
    from src.agents.country import Countries


class Household(Agent):

    def __init__(self, pid, od_point, name, long, lat, population, sector_consumption):
        super().__init__(
            agent_type="household",
            name=name,
            pid=pid,
            od_point=od_point,
            long=long,
            lat=lat
        )
        # Parameters depending on data
        self.sector_consumption = sector_consumption
        self.population = population
        # Parameters depending on network
        self.purchase_plan = {}
        self.retailers = {}
        # Variables reset and updated at each time step
        self.consumption_per_retailer = {}
        self.tot_consumption = 0
        self.consumption_per_sector = {}
        self.consumption_loss_per_sector = {}
        self.spending_per_retailer = {}
        self.tot_spending = 0
        self.spending_per_sector = {}
        self.extra_spending_per_sector = {}
        # Cumulated variables reset at beginning and updated at each time step
        self.consumption_loss = 0
        self.extra_spending = 0

    def reset_variables(self):
        self.consumption_per_retailer = {}
        self.tot_consumption = 0
        self.spending_per_retailer = {}
        self.tot_spending = 0
        self.extra_spending = 0
        self.consumption_loss = 0
        self.extra_spending_per_sector = {}
        self.consumption_loss_per_sector = {}

    def reset_indicators(self):
        self.consumption_per_retailer = {}
        self.tot_consumption = 0
        self.spending_per_retailer = {}
        self.tot_spending = 0
        self.extra_spending = 0
        self.consumption_loss = 0
        self.extra_spending_per_sector = {}
        self.consumption_loss_per_sector = {}

    def update_indicator(self, quantity_delivered: float, price: float, commercial_link: "CommercialLink"):
        super().update_indicator(quantity_delivered, price, commercial_link)
        self.consumption_per_retailer[commercial_link.supplier_id] = quantity_delivered
        self.tot_consumption += quantity_delivered
        self.spending_per_retailer[commercial_link.supplier_id] = quantity_delivered * price
        self.tot_spending += quantity_delivered * price
        new_extra_spending = quantity_delivered * (price - commercial_link.eq_price)
        self.extra_spending += new_extra_spending
        add_or_increment_dict_key(self.extra_spending_per_sector, commercial_link.product, new_extra_spending)
        new_consumption_loss = (self.purchase_plan[commercial_link.supplier_id] - quantity_delivered) \
                               * commercial_link.eq_price
        self.consumption_loss += new_consumption_loss
        add_or_increment_dict_key(self.consumption_loss_per_sector, commercial_link.product, new_consumption_loss)

    def initialize_var_on_purchase_plan(self):
        if len(self.purchase_plan) == 0:
            logging.warning("Households initialize variables based on purchase plan, but it is empty.")

        self.consumption_per_retailer = self.purchase_plan
        self.tot_consumption = sum(list(self.purchase_plan.values()))
        self.consumption_loss_per_sector = {sector: 0 for sector in self.purchase_plan.keys()}
        self.spending_per_retailer = self.consumption_per_retailer
        self.tot_spending = self.tot_consumption
        self.extra_spending_per_sector = {sector: 0 for sector in self.purchase_plan.keys()}

    def identify_suppliers(self, sector: str, firms: "Firms", countries: "Countries",
                           nb_suppliers_per_input: float, weight_localization: float, force_local: bool,
                           firm_data_type: str):
        if firm_data_type == "mrio":
            # if len(sector_id) == 3:  # case of countries
            if import_label in sector:  # case of countries
                supplier_type = "country"
                selected_supplier_ids = [sector[:3]]  # for countries, the id is extracted from the name
                supplier_weights = [1]

            else:  # case of firms
                supplier_type = "firm"
                potential_supplier_pids = [pid for pid, firm in firms.items() if firm.sector == sector]
                if len(potential_supplier_pids) == 0:
                    raise ValueError(f"{self.id_str().capitalize()}: there should be one supplier for {sector}")
                # Choose based on importance
                prob_to_be_selected = np.array(rescale_values([firms[firm_pid].importance for firm_pid in
                                                               potential_supplier_pids]))
                prob_to_be_selected /= prob_to_be_selected.sum()
                selected_supplier_ids = np.random.choice(potential_supplier_pids,
                                                         p=prob_to_be_selected, size=1,
                                                         replace=False).tolist()
                supplier_weights = [1]

        else:
            if sector == "IMP":
                supplier_type = "country"
                # Identify countries as suppliers if the corresponding sector does export
                importance_threshold = 1e-6
                potential_suppliers = [
                    pid
                    for pid, country in countries.items()
                    if country.supply_importance > importance_threshold
                ]
                importance_of_each = [
                    country.supply_importance
                    for country in countries.values()
                    if country.supply_importance > importance_threshold
                ]
                prob_to_be_selected = np.array(importance_of_each)
                prob_to_be_selected /= prob_to_be_selected.sum()

            # For the other types of inputs, identify the domestic suppliers, and
            # calculate their probability to be chosen, based on distance and importance
            else:
                supplier_type = "firm"
                potential_suppliers = [pid for pid, firm in firms.items() if firm.sector == sector]
                if len(potential_suppliers) == 0:
                    raise ValueError(f"{self.id_str().capitalize()}: no supplier for input {sector}")
                if force_local:
                    potential_local_suppliers = [firm_id for firm_id in potential_suppliers
                                                 if firms[firm_id].od_point == self.od_point]
                    if len(potential_local_suppliers) > 0:
                        potential_suppliers = potential_local_suppliers
                    else:
                        pass
                        # logging.debug(f"{self.id_str().capitalize()}: no local supplier for input {sector_id}")
                distance_to_each = rescale_values([
                    calculate_distance_between_agents(self, firms[firm_id])
                    for firm_id in potential_suppliers
                ])

                importance_of_each = rescale_values([firms[firm_id].importance for firm_id in potential_suppliers])

                prob_to_be_selected = np.array(importance_of_each) / (np.array(distance_to_each) ** weight_localization)
                prob_to_be_selected /= prob_to_be_selected.sum()

            # Determine the number of supplier(s) to select. 1 or 2.
            # nb_suppliers_to_choose = determine_nb_suppliers(nb_suppliers_per_input)
            nb_suppliers_to_choose = 1  # for households, always 1

            # Select the supplier(s). It there is 2 suppliers, then we generate
            # random weight. It determines how much is bought from each supplier.
            selected_supplier_ids = np.random.choice(potential_suppliers,
                                                     p=prob_to_be_selected, size=nb_suppliers_to_choose,
                                                     replace=False).tolist()
            index_map = {supplier_id: position for position, supplier_id in enumerate(potential_suppliers)}
            selected_positions = [index_map[supplier_id] for supplier_id in selected_supplier_ids]
            selected_prob = [prob_to_be_selected[position] for position in selected_positions]
            supplier_weights = generate_weights(nb_suppliers_to_choose, selected_prob)

        return supplier_type, selected_supplier_ids, supplier_weights

    def select_suppliers(self, graph: "ScNetwork", firms: "Firms", countries: "Countries",
                         nb_retailers: float, force_local: bool,
                         weight_localization: float, firm_data_type: str):
        # print(f"{self.id_str()}: consumption {self.sector_consumption}")
        for sector, amount in self.sector_consumption.items():
            supplier_type, retailers, retailer_weights = self.identify_suppliers(sector, firms, countries,
                                                                                 nb_retailers,
                                                                                 weight_localization,
                                                                                 force_local,
                                                                                 firm_data_type)

            # For each of them, create commercial link
            for retailer_id in retailers:
                # Retrieve the appropriate supplier object from the id
                # If it is a country we get it from the country list
                # If it is a firm we get it from the firm list
                if supplier_type == "country":
                    supplier_object = countries[retailer_id]
                    link_category = 'import_B2C'
                    product_type = "imports"
                else:
                    supplier_object = firms[retailer_id]
                    link_category = 'domestic_B2C'
                    product_type = firms[retailer_id].sector_type

                # For each retailer, create an edge in the economic network
                graph.add_edge(supplier_object, self,
                               object=CommercialLink(
                                   pid=str(retailer_id) + '->' + str(self.pid),
                                   product=sector,
                                   product_type=product_type,
                                   category=link_category,
                                   supplier_id=retailer_id,
                                   buyer_id=self.pid)
                               )
                # Associate a weight in the commercial link, the household's purchase plan & retailer list, in the retailer's client list
                weight = retailer_weights.pop()
                graph[supplier_object][self]['weight'] = weight
                self.purchase_plan[retailer_id] = weight * self.sector_consumption[sector]
                self.retailers[retailer_id] = {'sector': sector, 'weight': weight}
                distance = calculate_distance_between_agents(self, supplier_object)
                supplier_object.clients[self.pid] = {
                    'sector': "households", 'share': 0, 'transport_share': 0, "distance": distance
                }  # The share of sales cannot be calculated now.

    def send_purchase_orders(self, graph):
        for edge in graph.in_edges(self):
            try:
                quantity_to_buy = self.purchase_plan[edge[0].pid]
            except KeyError:
                print("Households: No purchase plan for supplier", edge[0].pid)
                quantity_to_buy = 0
            graph[edge[0]][self]['object'].order = quantity_to_buy

    def select_supplier_from_list(self, firm_list: "Firms",
                                  nb_suppliers_to_choose: int, potential_firm_ids: list,
                                  distance: bool, importance: bool, weight_localization: float,
                                  force_same_odpoint=False):
        # reduce firm to choose to local ones
        if force_same_odpoint:
            same_odpoint_firms = [
                firm_id
                for firm_id in potential_firm_ids
                if firm_list[firm_id].odpoint == self.od_point
            ]
            if len(same_odpoint_firms) > 0:
                potential_firm_ids = same_odpoint_firms
            #     logging.info('retailer available locally at od_point '+str(agent.od_point)+
            #         " for "+firms[potential_firm_ids[0]].sector)
            # else:
            #     logging.warning('force_same_odpoint but no retailer available at od_point '+str(agent.od_point)+
            #         " for "+firms[potential_firm_ids[0]].sector)

        # distance weight
        if distance:
            distance_to_each = rescale_values([
                calculate_distance_between_agents(self, firm_list[firm_id])
                for firm_id in potential_firm_ids
            ])
            distance_weight = 1 / (np.array(distance_to_each) ** weight_localization)
        else:
            distance_weight = np.ones(len(potential_firm_ids))

        # importance weight
        if importance:
            importance_of_each = rescale_values([firm_list[firm_id].importance for firm_id in potential_firm_ids])
            importance_weight = np.array(importance_of_each)
        else:
            importance_weight = np.ones(len(potential_firm_ids))

        # create weight vector based on choice
        prob_to_be_selected = distance_weight * importance_weight
        prob_to_be_selected /= prob_to_be_selected.sum()

        # perform the random choice
        selected_supplier_ids = np.random.choice(
            potential_firm_ids,
            p=prob_to_be_selected,
            size=nb_suppliers_to_choose,
            replace=False
        ).tolist()
        # Choose weight if there are multiple suppliers
        index_map = {supplier_id: position for position, supplier_id in enumerate(potential_firm_ids)}
        selected_positions = [index_map[supplier_id] for supplier_id in selected_supplier_ids]
        selected_prob = [prob_to_be_selected[position] for position in selected_positions]
        supplier_weights = generate_weights(nb_suppliers_to_choose, importance_of_each=selected_prob)

        # return
        return selected_supplier_ids, supplier_weights


class Households(Agents):
    pass



from typing import TYPE_CHECKING

import logging
from pathlib import Path

import geopandas
import pandas
import pandas as pd
import geopandas as gpd
from pandas import Series

if TYPE_CHECKING:
    from src.agents.firm import Firms
    from src.agents.country import Countries


def filter_sector(sector_table, cutoff_sector_output, cutoff_sector_demand,
                  combine_sector_cutoff='and', sectors_to_include="all", sectors_to_exclude=None):
    """Filter the sector table to sector whose output and/or final demand is larger than cutoff values
    In addition to filters, we can force to exclude or include some sectors

    Parameters
    ----------
    sector_table : pandas.DataFrame
        Sector table
    cutoff_sector_output : dictionary
        Cutoff parameters for selecting the sectors based on output
        If type="percentage", the sector's output divided by all sectors' output is used
        If type="absolute", the sector's absolute output, in USD, is used
        If type="relative_to_average", the cutoff used is (cutoff value) * (country's total output) / (nb sectors)
    cutoff_sector_demand : dictionary
        Cutoff value for selecting the sectors based on final demand
        If type="percentage", the sector's final demand divided by all sectors' output is used
        If type="absolute", the sector's absolute output, in USD, is used
    combine_sector_cutoff: "and", "or"
        If 'and', select sectors that pass both the output and demand cutoff
        If 'or', select sectors that pass either the output or demand cutoff
    sectors_to_include : list of string or 'all'
        list of the sectors preselected by the user. Default to "all"
    sectors_to_exclude : list of string or None
        list of the sectors pre-eliminated by the user. Default to None

    Returns
    -------
    list of filtered sectors
    """
    # Select sectors based on output
    filtered_sectors_output = apply_sector_filter(sector_table, 'output', cutoff_sector_output)
    filtered_sectors_demand = apply_sector_filter(sector_table, 'final_demand', cutoff_sector_demand)

    # Merge both list
    if combine_sector_cutoff == 'and':
        filtered_sectors = list(set(filtered_sectors_output) & set(filtered_sectors_demand))
    elif combine_sector_cutoff == 'or':
        filtered_sectors = list(set(filtered_sectors_output + filtered_sectors_demand))
    else:
        raise ValueError("'combine_sector_cutoff' should be 'and' or 'or'")

        # Force to include some sector
    if isinstance(sectors_to_include, list):
        if len(set(sectors_to_include) - set(filtered_sectors)) > 0:
            selected_but_filtered_out_sectors = list(set(sectors_to_include) - set(filtered_sectors))
            logging.info("The following sectors were specifically selected but were filtered out" +
                         str(selected_but_filtered_out_sectors))
        filtered_sectors = list(set(sectors_to_include) & set(filtered_sectors))

    # Force to exclude some sectors
    if sectors_to_exclude:
        filtered_sectors = [sector for sector in filtered_sectors if sector not in sectors_to_exclude]
    if len(filtered_sectors) == 0:
        raise ValueError("We excluded all sectors")

    # Sort list
    filtered_sectors.sort()
    return filtered_sectors


def apply_sector_filter(sector_table, filter_column, cut_off_dic):
    """Filter the sector_table using the filter_column
    The way to cut_off is defined in cut_off_dic

    sector_table : pandas.DataFrame
        Sector table
    filter_column : string
        'output' or 'final_demand'
    cut_off_dic : dictionary
        Cutoff parameters for selecting the sectors based on output
        If type="percentage", the sector's filter_column divided by all sectors' output is used
        If type="absolute", the sector's absolute filter_column is used
        If type="relative_to_average", the cutoff used is (cutoff value) * (total filter_column) / (nb sectors)
    """
    sector_table_no_import = sector_table[sector_table['sector'] != "IMP"]

    if cut_off_dic['type'] == "percentage":
        rel_output = sector_table_no_import[filter_column] / sector_table_no_import['output'].sum()
        filtered_sectors = sector_table_no_import.loc[
            rel_output > cut_off_dic['value'],
            "sector"
        ].tolist()
    elif cut_off_dic['type'] == "absolute":
        filtered_sectors = sector_table_no_import.loc[
            sector_table_no_import[filter_column] > cut_off_dic['value'],
            "sector"
        ].tolist()
    elif cut_off_dic['type'] == "relative_to_average":
        cutoff = cut_off_dic['value'] \
                 * sector_table_no_import[filter_column].sum() \
                 / sector_table_no_import.shape[0]
        filtered_sectors = sector_table_no_import.loc[
            sector_table_no_import['output'] > cutoff,
            "sector"
        ].tolist()
    else:
        raise ValueError("cutoff type should be 'percentage', 'absolute', or 'relative_to_average'")
    if len(filtered_sectors) == 0:
        raise ValueError("The output cutoff value is so high that it filtered out all sectors")
    return filtered_sectors


def get_closest_road_nodes(regions: pd.Series,
                           transport_nodes: geopandas.GeoDataFrame, filepath_region_table: Path) -> pd.Series:
    region_table = gpd.read_file(filepath_region_table)
    dic_region_to_points = region_table.set_index('region')['geometry'].to_dict()
    road_nodes = transport_nodes[transport_nodes['type'] == "roads"]
    dic_region_to_road_node_id = {
        region: road_nodes.loc[get_index_closest_point(point, road_nodes), 'id']
        for region, point in dic_region_to_points.items()
    }
    closest_road_nodes = regions.map(dic_region_to_road_node_id)
    if closest_road_nodes.isnull().sum() > 0:
        logging.warning(f"{closest_road_nodes.isnull().sum()} regions not found")
        raise KeyError(f"{closest_road_nodes.isnull().sum()} regions not found: "
                       f"{regions[closest_road_nodes.isnull()].to_list()}")
    return closest_road_nodes


def get_long_lat(nodes_ids: pd.Series, transport_nodes: geopandas.GeoDataFrame) -> dict[str, Series]:
    od_point_table = transport_nodes[transport_nodes['id'].isin(nodes_ids)].copy()
    od_point_table['long'] = od_point_table.geometry.x
    od_point_table['lat'] = od_point_table.geometry.y
    road_node_id_to_long_lat = od_point_table.set_index('id')[['long', 'lat']]
    return {
        'long': nodes_ids.map(road_node_id_to_long_lat['long']),
        'lat': nodes_ids.map(road_node_id_to_long_lat['long'])
    }


def get_index_closest_point(point, df_with_points):
    """Given a point it finds the index of the closest points in a Point GeoDataFrame.

    Parameters
    ----------
    point: shapely.Point
        Point object of which we want to find the closest point
    df_with_points: geopandas.GeoDataFrame
        containing the points among which we want to find the
        one that is the closest to point

    Returns
    -------
    type depends on the index data type of df_with_points
        index object of the closest point in df_with_points
    """
    distance_list = [point.distance(item) for item in df_with_points['geometry'].tolist()]
    return df_with_points.index[distance_list.index(min(distance_list))]


def extract_final_list_of_sector(firms: "Firms"):
    n = len(firms)
    present_sectors = list(set([firm.main_sector for firm in firms.values()]))
    present_sectors.sort()
    flow_types_to_export = present_sectors + ['domestic_B2C', 'domestic_B2B', 'transit', 'import',
                                              'import_B2C', 'export', 'total']
    logging.info('Firm_list created, size is: ' + str(n))
    logging.info('Sectors present are: ' + str(present_sectors))
    return n, present_sectors, flow_types_to_export


def load_ton_usd_equivalence(sector_table: pd.DataFrame, firm_table: pd.DataFrame,
                             firms: "Firms", countries: "Countries"):
    """Load equivalence between usd and ton

    It updates the firms and countries.
    It updates the 'usd_per_ton' attribute of firms, based on their sector.
    It updates the 'usd_per_ton' attribute of countries, it gives the average.
    Note that this will be applied only to goods that are delivered by those agents.

    sector_table : pandas.DataFrame
        Sector table
    firms : list(Firm objects)
        list of firms
    countries : list(Country objects)
        list of countries
    """
    sector_to_usd_per_ton = sector_table.set_index('sector')['usd_per_ton']
    firm_table['usd_per_ton'] = firm_table['sector'].map(sector_to_usd_per_ton)
    for firm in firms.values():
        firm.usd_per_ton = sector_to_usd_per_ton[firm.sector]

    for country in countries.values():
        country.usd_per_ton = sector_to_usd_per_ton['IMP']



import logging
import pickle

from src.paths import TMP_FOLDER


def generate_cache_parameters_from_command_line_argument(arguments: list[str]):
    # Generate cache parameters
    cache_parameters: dict[str, bool] = {
        "transport_network": False,
        "agents": False,
        "sc_network": False,
        "logistic_routes": False
    }
    if len(arguments) > 2:
        accepted_script_arguments: list[str] = [
            'same_transport_network_new_agents',
            'same_agents_new_sc_network',
            'same_sc_network_new_logistic_routes',
            'same_logistic_routes',
            'same_agents_new_transport_network',
            'same_sc_network_new_transport_network',
            'new_agents_same_all'
        ]
        argument = arguments[2]
        if argument not in accepted_script_arguments:
            raise ValueError(f"Argument {argument} is not valid.\
                Possible values are: " + ','.join(accepted_script_arguments))
        cache_parameters: dict[str, bool] = {
            "transport_network": False,
            "agents": False,
            "sc_network": False,
            "logistic_routes": False
        }
        if argument == accepted_script_arguments[0]:
            cache_parameters['transport_network'] = True
        if argument == accepted_script_arguments[1]:
            cache_parameters['transport_network'] = True
            cache_parameters['agents'] = True
        if argument == accepted_script_arguments[2]:
            cache_parameters['transport_network'] = True
            cache_parameters['agents'] = True
            cache_parameters['sc_network'] = True
        if argument == accepted_script_arguments[3]:
            cache_parameters['transport_network'] = True
            cache_parameters['agents'] = True
            cache_parameters['sc_network'] = True
            cache_parameters['logistic_routes'] = True
        if argument == accepted_script_arguments[4]:
            cache_parameters['transport_network'] = False
            cache_parameters['agents'] = True
            cache_parameters['sc_network'] = False
            cache_parameters['logistic_routes'] = False
        if argument == accepted_script_arguments[5]:
            cache_parameters['transport_network'] = False
            cache_parameters['agents'] = True
            cache_parameters['sc_network'] = True
            cache_parameters['logistic_routes'] = False
        if argument == accepted_script_arguments[6]:
            cache_parameters['transport_network'] = True
            cache_parameters['agents'] = False
            cache_parameters['sc_network'] = True
            cache_parameters['logistic_routes'] = True
    return cache_parameters


def cache_agent_data(data_dic):
    pickle_filename = TMP_FOLDER / 'firms_households_countries_pickle'
    pickle.dump(data_dic, open(pickle_filename, 'wb'))
    logging.info(f'Firms, households, and countries saved in tmp folder: {pickle_filename}')


def cache_transport_network(data_dic):
    pickle_filename = TMP_FOLDER / 'transport_network_pickle'
    pickle.dump(data_dic, open(pickle_filename, 'wb'))
    logging.info(f'Transport network saved in tmp folder: {pickle_filename}')


def cache_sc_network(data_dic):
    pickle_filename = TMP_FOLDER / 'supply_chain_pickle'
    pickle.dump(data_dic, open(pickle_filename, 'wb'))
    logging.info(f'Supply chain saved in tmp folder: {pickle_filename}')


def cache_logistic_routes(data_dic):
    pickle_filename = TMP_FOLDER / 'logistic_routes_pickle'
    pickle.dump(data_dic, open(pickle_filename, 'wb'))
    logging.info(f'Logistics routes saved in tmp folder: {pickle_filename}')


def load_cached_agent_data():
    pickle_filename = TMP_FOLDER / 'firms_households_countries_pickle'
    tmp_data = pickle.load(open(pickle_filename, 'rb'))
    loaded_sector_table = tmp_data['sector_table']
    # loaded_present_sectors = tmp_data['present_sectors']
    # loaded_flow_types_to_export = tmp_data['flow_types_to_export']
    loaded_firm_table = tmp_data['firm_table']
    loaded_household_table = tmp_data['household_table']
    loaded_firms = tmp_data['firms']
    loaded_households = tmp_data['households']
    loaded_countries = tmp_data['countries']
    logging.info('Firms, households, and countries generated from temp file.')
    logging.info("Nb firms: " + str(len(loaded_firms)))
    logging.info("Nb households: " + str(len(loaded_households)))
    logging.info("Nb countries: " + str(len(loaded_countries)))
    return loaded_sector_table, loaded_firms, loaded_firm_table, \
        loaded_households, loaded_household_table, loaded_countries


def load_cached_transaction_table():
    pickle_filename = TMP_FOLDER / 'firms_households_countries_pickle'
    tmp_data = pickle.load(open(pickle_filename, 'rb'))
    loaded_transaction_table = tmp_data['transaction_table']
    return loaded_transaction_table


def load_cached_transport_network():
    pickle_filename = TMP_FOLDER / 'transport_network_pickle'
    tmp_data = pickle.load(open(pickle_filename, 'rb'))
    loaded_transport_network = tmp_data['transport_network']
    loaded_transport_nodes = tmp_data['transport_nodes']
    loaded_transport_edges = tmp_data['transport_edges']
    logging.info('Transport network generated from temp file.')
    return loaded_transport_network, loaded_transport_nodes, loaded_transport_edges


def load_cached_sc_network():
    pickle_filename = TMP_FOLDER / 'supply_chain_pickle'
    tmp_data = pickle.load(open(pickle_filename, 'rb'))
    loaded_sc_network = tmp_data['supply_chain_network']
    loaded_firms = tmp_data['firms']
    loaded_households = tmp_data['households']
    loaded_countries = tmp_data['countries']
    logging.info('Supply chain generated from temp file.')
    return loaded_sc_network, loaded_firms, loaded_households, loaded_countries


def load_cached_logistic_routes():
    pickle_filename = TMP_FOLDER / 'logistic_routes_pickle'
    tmp_data = pickle.load(open(pickle_filename, 'rb'))
    loaded_sc_network = tmp_data['supply_chain_network']
    loaded_transport_network = tmp_data['transport_network']
    loaded_firms = tmp_data['firms']
    loaded_households = tmp_data['households']
    loaded_countries = tmp_data['countries']
    logging.info('Logistic routes generated from temp file.')
    return loaded_sc_network, loaded_transport_network, loaded_firms, loaded_households, loaded_countries



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import geopandas as gpd
import logging

from src.model.basic_functions import calculate_distance_between_agents, rescale_values, \
    generate_weights_from_list
from src.agents.agent import Agent, Agents
from src.network.commercial_link import CommercialLink

if TYPE_CHECKING:
    from src.network.transport_network import TransportNetwork
    from src.network.sc_network import ScNetwork


class Country(Agent):

    def __init__(self, pid=None, qty_sold=None, qty_purchased=None, od_point=None, long=None, lat=None,
                 purchase_plan=None, transit_from=None, transit_to=None, supply_importance=None,
                 usd_per_ton=None, sector="IMP"):
        # Intrinsic parameters
        super().__init__(
            agent_type="country",
            pid=pid,
            od_point=od_point,
            long=long,
            lat=lat
        )
        self.sector = sector

        # Parameter based on data
        self.usd_per_ton = usd_per_ton
        # self.entry_points = entry_points or []
        self.transit_from = transit_from or {}
        self.transit_to = transit_to or {}
        self.supply_importance = supply_importance

        # Parameters depending on supplier-buyer network
        self.clients = {}
        self.purchase_plan = purchase_plan or {}
        self.qty_sold = qty_sold or {}
        self.qty_purchased = qty_purchased or {}
        self.qty_purchased_perfirm = {}

        # Variable
        self.generalized_transport_cost = 0
        self.usd_transported = 0
        self.tons_transported = 0
        self.tonkm_transported = 0
        self.extra_spending = 0
        self.consumption_loss = 0

    def reset_variables(self):
        self.generalized_transport_cost = 0
        self.usd_transported = 0
        self.tons_transported = 0
        self.tonkm_transported = 0
        self.extra_spending = 0
        self.consumption_loss = 0

    def reset_indicators(self):
        self.extra_spending = 0
        self.consumption_loss = 0

    def update_indicator(self, quantity_delivered: float, price: float, commercial_link: "CommercialLink"):
        super().update_indicator(quantity_delivered, price, commercial_link)
        self.extra_spending += quantity_delivered * (price - commercial_link.eq_price)
        self.consumption_loss += commercial_link.delivery - quantity_delivered

    def create_transit_links(self, graph, countries):
        for selling_country_pid, quantity in self.transit_from.items():
            selling_country_object = [country for pid, country in countries.items() if pid == selling_country_pid][0]
            graph.add_edge(selling_country_object, self,
                           object=CommercialLink(
                               pid=str(selling_country_pid) + '->' + str(self.pid),
                               product='transit',
                               product_type="transit",  # suppose that transit type are non service, material stuff
                               category="transit",
                               supplier_id=selling_country_pid,
                               buyer_id=self.pid))
            graph[selling_country_object][self]['weight'] = 1
            self.purchase_plan[selling_country_pid] = quantity
            selling_country_object.clients[self.pid] = {'sector': self.pid, 'share': 0, 'transport_share': 0}

    def select_suppliers(self, graph, firms, country_list,
                         sector_table: pd.DataFrame, transport_nodes: gpd.GeoDataFrame):
        # Select other country as supplier: transit flows
        self.create_transit_links(graph, country_list)

        # Identify firms from each sector
        dic_sector_to_firm_id = identify_firms_in_each_sector(firms)
        share_exporting_firms = sector_table.set_index('sector')['share_exporting_firms'].to_dict()
        # Identify od_points which exports (optional)
        export_od_points = transport_nodes.dropna(subset=['special'])
        export_od_points = export_od_points.loc[export_od_points['special'].str.contains("export"), "id"].tolist()
        # Identify sectors to buy from
        present_sectors = list(set(list(dic_sector_to_firm_id.keys())))
        sectors_to_buy_from = list(self.qty_purchased.keys())
        present_sectors_to_buy_from = list(set(present_sectors) & set(sectors_to_buy_from))
        # For each one of these sectors, select suppliers
        supplier_selection_mode = {
            "importance_export": {
                "export_od_points": export_od_points,
                "bonus": 10
                # give more weight to firms located in transport node identified as "export points" (e.g., SEZs)
            }
        }
        for sector in present_sectors_to_buy_from:  # only select suppliers from sectors that are present
            # Identify potential suppliers
            potential_supplier_pid = dic_sector_to_firm_id[sector]
            # Evaluate how much to select
            if sector not in share_exporting_firms:  # case of mrio
                nb_suppliers_to_select = 1
            else:  # otherwise we use the % of the sector table to cal
                nb_suppliers_to_select = max(1, round(len(potential_supplier_pid) * share_exporting_firms[sector]))
            if nb_suppliers_to_select > len(potential_supplier_pid):
                logging.warning(f"The number of supplier to select {nb_suppliers_to_select} "
                                f"is larger than the number of potential supplier {len(potential_supplier_pid)} "
                                f"(share_exporting_firms: {share_exporting_firms[sector]})")
                # Select supplier and weights
            selected_supplier_ids, supplier_weights = determine_suppliers_and_weights(
                potential_supplier_pid,
                nb_suppliers_to_select,
                firms,
                mode=supplier_selection_mode
            )
            # Materialize the link
            for supplier_id in selected_supplier_ids:
                # For each supplier, create an edge in the economic network
                graph.add_edge(firms[supplier_id], self,
                               object=CommercialLink(
                                   pid=str(supplier_id) + '->' + str(self.pid),
                                   product=sector,
                                   product_type=firms[supplier_id].sector_type,
                                   category="export",
                                   supplier_id=supplier_id,
                                   buyer_id=self.pid))
                # Associate a weight
                weight = supplier_weights.pop(0)
                graph[firms[supplier_id]][self]['weight'] = weight
                # Households save the name of the retailer, its sector, its weight, and adds it to its purchase plan
                self.qty_purchased_perfirm[supplier_id] = {
                    'sector': sector,
                    'weight': weight,
                    'amount': self.qty_purchased[sector] * weight
                }
                self.purchase_plan[supplier_id] = self.qty_purchased[sector] * weight
                # The supplier saves the fact that it exports to this country.
                # The share of sales cannot be calculated now, we put 0 for the moment
                distance = calculate_distance_between_agents(self, firms[supplier_id])
                firms[supplier_id].clients[self.pid] = {
                    'sector': self.pid, 'share': 0, 'transport_share': 0, 'distance': distance
                }

    def send_purchase_orders(self, graph):
        for edge in graph.in_edges(self):
            try:
                quantity_to_buy = self.purchase_plan[edge[0].pid]
            except KeyError:
                print("Country " + self.pid + ": No purchase plan for supplier", edge[0].pid)
                quantity_to_buy = 0
            graph[edge[0]][self]['object'].order = quantity_to_buy

    def deliver_products(self, graph: "ScNetwork", transport_network: "TransportNetwork",
                         sectors_no_transport_network: list[str], rationing_mode: str, monetary_units_in_model: str,
                         cost_repercussion_mode: str, price_increase_threshold: float, capacity_constraint: bool,
                         transport_cost_noise_level: float):
        """ The quantity to be delivered is the quantity that was ordered (no rationing takes place)

        Parameters
        ----------
        transport_cost_noise_level
        cost_repercussion_mode
        capacity_constraint
        monetary_units_in_model
        rationing_mode
        sectors_no_transport_network
        transport_network
        graph
        price_increase_threshold
        """
        self.generalized_transport_cost = 0
        self.usd_transported = 0
        self.tons_transported = 0
        self.tonkm_transported = 0
        self.qty_sold = 0
        for edge in graph.out_edges(self):
            if graph[self][edge[1]]['object'].order == 0:
                logging.debug(f"{self.id_str()} - {edge[1].id_str()} is my client but did not order")
                continue
            graph[self][edge[1]]['object'].delivery = graph[self][edge[1]]['object'].order
            graph[self][edge[1]]['object'].delivery_in_tons = \
                Country.transformUSD_to_tons(graph[self][edge[1]]['object'].order, monetary_units_in_model,
                                             self.usd_per_ton)

            explicit_service_firm = True
            if explicit_service_firm:
                # If send services, no use of transport network
                if graph[self][edge[1]]['object'].product_type in sectors_no_transport_network:
                    graph[self][edge[1]]['object'].price = graph[self][edge[1]]['object'].eq_price
                    self.qty_sold += graph[self][edge[1]]['object'].delivery
                # Otherwise, send shipment through transportation network
                else:
                    self.send_shipment(graph[self][edge[1]]['object'], transport_network, monetary_units_in_model,
                                       cost_repercussion_mode, price_increase_threshold, capacity_constraint,
                                       transport_cost_noise_level)
            else:
                if (edge[1].od_point != -1):  # to non-service firms, send shipment through transportation network
                    self.send_shipment(graph[self][edge[1]]['object'], transport_network, monetary_units_in_model,
                                       cost_repercussion_mode, price_increase_threshold, capacity_constraint,
                                       transport_cost_noise_level)
                else:  # if it sends to service firms, nothing to do. price is equilibrium price
                    graph[self][edge[1]]['object'].price = graph[self][edge[1]]['object'].eq_price
                    self.qty_sold += graph[self][edge[1]]['object'].delivery

    def send_shipment(self, commercial_link: "CommercialLink", transport_network: "TransportNetwork",
                      monetary_units_in_model: str, cost_repercussion_mode: str, price_increase_threshold: float,
                      capacity_constraint: bool, transport_cost_noise_level: float):

        if commercial_link.delivery_in_tons == 0:
            print("delivery", commercial_link.delivery)
            print("supplier_id", commercial_link.supplier_id)
            print("buyer_id", commercial_link.buyer_id)

        monetary_unit_factor = {
            "mUSD": 1e6,
            "kUSD": 1e3,
            "USD": 1
        }
        factor = monetary_unit_factor[monetary_units_in_model]
        """Only apply to B2B flows 
        """
        if len(commercial_link.route) == 0:
            raise ValueError("Country " + str(self.pid) +
                             ": commercial link " + str(commercial_link.pid) +
                             " is not associated to any route, I cannot send any shipment to client " +
                             str(commercial_link.pid))

        if self.check_route_availability(commercial_link, transport_network, 'main') == 'available':
            # If the normal route is available, we can send the shipment as usual and pay the usual price
            commercial_link.current_route = 'main'
            commercial_link.price = commercial_link.eq_price
            transport_network.transport_shipment(commercial_link, capacity_constraint)

            self.generalized_transport_cost += commercial_link.route_time_cost \
                                               + commercial_link.delivery_in_tons * commercial_link.route_cost_per_ton
            self.usd_transported += commercial_link.delivery
            self.tons_transported += commercial_link.delivery_in_tons
            self.tonkm_transported += commercial_link.delivery_in_tons * commercial_link.route_length
            self.qty_sold += commercial_link.delivery
            return 0

        # If there is a disruption, we try the alternative route, if there is any
        if (len(commercial_link.alternative_route) > 0) & \
                (self.check_route_availability(commercial_link, transport_network, 'alternative') == 'available'):
            commercial_link.current_route = 'alternative'
            route = commercial_link.alternative_route
        # Otherwise we have to find a new one
        else:
            origin_node = self.od_point
            destination_node = commercial_link.route[-1][0]
            route = self.choose_route(
                transport_network=transport_network.get_undisrupted_network(),
                origin_node=origin_node,
                destination_node=destination_node,
                capacity_constraint=capacity_constraint,
                transport_cost_noise_level=transport_cost_noise_level
            )
            # We evaluate the cost of this new route
            if route is not None:
                commercial_link.store_route_information(
                    route=route,
                    main_or_alternative="alternative"
                )

        # If the alternative route is available, or if we discovered one, we proceed
        if route is not None:
            commercial_link.current_route = 'alternative'
            # Calculate contribution to generalized transport cost, to usd/tons/tonkms transported
            self.generalized_transport_cost += commercial_link.alternative_route_time_cost \
                                               + commercial_link.delivery_in_tons * commercial_link.alternative_route_cost_per_ton
            self.usd_transported += commercial_link.delivery
            self.tons_transported += commercial_link.delivery_in_tons
            self.tonkm_transported += commercial_link.delivery_in_tons * commercial_link.alternative_route_length
            self.qty_sold += commercial_link.delivery

            if cost_repercussion_mode == "type1":  # relative cost change with actual bill
                # Calculate relative increase in routing cost
                new_transport_bill = commercial_link.delivery_in_tons * commercial_link.alternative_route_cost_per_ton
                normal_transport_bill = commercial_link.delivery_in_tons * commercial_link.route_cost_per_ton
                relative_cost_change = max(new_transport_bill - normal_transport_bill, 0) / normal_transport_bill
                # print(
                #     self.pid,
                #     commercial_link.delivery_in_tons,
                #     commercial_link.route_cost_per_ton,
                #     commercial_link.alternative_route_cost_per_ton,
                #     relative_cost_change
                # )
                # If switched transport mode, add switching cost
                # switching_cost = 0.5
                # if commercial_link.alternative_route_mode != commercial_link.route_mode:
                #     relative_cost_change = relative_cost_change + switching_cost
                # Translate that into an increase in transport costs in the balance sheet
                relative_price_change_transport = 0.2 * relative_cost_change
                total_relative_price_change = relative_price_change_transport
                commercial_link.price = commercial_link.eq_price * (1 + total_relative_price_change)

            elif cost_repercussion_mode == "type2":  # actual repercussion de la bill
                added_costUSD_per_ton = max(
                    commercial_link.alternative_route_cost_per_ton - commercial_link.route_cost_per_ton, 0)
                added_costUSD_per_mUSD = added_costUSD_per_ton / (self.usd_per_ton / factor)
                added_costmUSD_per_mUSD = added_costUSD_per_mUSD / factor
                commercial_link.price = commercial_link.eq_price + added_costmUSD_per_mUSD
                relative_price_change_transport = commercial_link.price / commercial_link.eq_price - 1

            elif cost_repercussion_mode == "type3":
                # We translate this real cost into transport cost
                relative_cost_change = (
                                               commercial_link.alternative_route_time_cost - commercial_link.route_time_cost) / commercial_link.route_time_cost
                relative_price_change_transport = 0.2 * relative_cost_change
                # With that, we deliver the shipment
                total_relative_price_change = relative_price_change_transport
                commercial_link.price = commercial_link.eq_price * (1 + total_relative_price_change)

            # If there is an alternative route but it is too expensive
            if relative_price_change_transport > price_increase_threshold:
                logging.info("Country " + str(self.pid) + ": found an alternative route to " +
                             str(commercial_link.buyer_id) + ", but it is costlier by " +
                             '{:.2f}'.format(100 * relative_price_change_transport) + "%, price would be " +
                             '{:.4f}'.format(commercial_link.price) + " instead of " +
                             '{:.4f}'.format(commercial_link.eq_price) +
                             ' so I decide not to send it now.'
                             )
                commercial_link.price = commercial_link.eq_price
                commercial_link.current_route = 'none'
                commercial_link.delivery = 0

            # If there is an alternative route which is not too expensive
            else:
                transport_network.transport_shipment(commercial_link, capacity_constraint)
                logging.info(f"{self.id_str().capitalize()}: found an alternative route "
                             f" client {commercial_link.buyer_id}, it is costlier by "
                             f"{100 * relative_price_change_transport:.0f}%, price is "
                             f"{commercial_link.price:.4f} instead of {commercial_link.eq_price:.4f}")

        # It there is no alternative route
        else:
            logging.info(f"{self.id_str().capitalize()}: because of disruption, there is " +
                         f"no route between me and client {commercial_link.buyer_id}")
            # We do not write how the input price would have changed
            commercial_link.price = commercial_link.eq_price
            # We do not pay the transporter, so we don't increment the transport cost

    def evaluate_commercial_balance(self, graph):
        exports = sum([graph[self][edge[1]]['object'].payment for edge in graph.out_edges(self)])
        imports = sum([graph[edge[0]][self]['object'].payment for edge in graph.in_edges(self)])
        print("Country " + self.pid + ": imports " + str(imports) + " from Tanzania and export " + str(
            exports) + " to Tanzania")


class Countries(Agents):
    pass


def identify_firms_in_each_sector(firms):
    firm_id_each_sector = pd.DataFrame({
        'firm': list(firms.keys()),
        'sector': [firm.sector for firm in firms.values()]})
    dic_sector_to_firmid = firm_id_each_sector \
        .groupby('sector')['firm'] \
        .apply(lambda x: list(x)) \
        .to_dict()
    return dic_sector_to_firmid


def determine_suppliers_and_weights(potential_supplier_pids,
                                    nb_selected_suppliers, firms, mode):
    # Get importance for each of them
    if "importance_export" in mode.keys():
        importance_of_each = rescale_values([
            firms[firm_pid].importance * mode['importance_export']['bonus']
            if firms[firm_pid].od_point in mode['importance_export']['export_od_points']
            else firms[firm_pid].importance
            for firm_pid in potential_supplier_pids
        ])
    elif "importance" in mode.keys():
        importance_of_each = rescale_values([
            firms[firm_pid].importance
            for firm_pid in potential_supplier_pids
        ])

    # Select supplier
    prob_to_be_selected = np.array(importance_of_each) / np.array(importance_of_each).sum()
    selected_supplier_ids = np.random.choice(potential_supplier_pids,
                                             p=prob_to_be_selected,
                                             size=nb_selected_suppliers,
                                             replace=False
                                             ).tolist()

    # Compute weights, based on importance only
    supplier_weights = generate_weights_from_list([
        firms[firm_pid].importance
        for firm_pid in selected_supplier_ids
    ])

    return selected_supplier_ids, supplier_weights



import logging

import pandas as pd

final_demand_label = "final_demand"
export_label = "Exports"
import_label = "Imports"
EPSILON = 1e-6

class Mrio(pd.DataFrame):
    _metadata = ['region_sectors', "region_sector_names", "regions", "sectors", "external_buying_countries",
                 "external_selling_countries", "region_households"]

    def __init__(self, *args, **kwargs):
        super(Mrio, self).__init__(*args, **kwargs)
        self.check_square_structure()
        self.region_sectors = [tup for tup in self.columns if tup[1] not in [final_demand_label, export_label]]
        self.region_sector_names = ['_'.join(tup) for tup in self.region_sectors]
        self.regions = list(set([tup[0] for tup in self.region_sectors]))
        self.sectors = list(set([tup[1] for tup in self.region_sectors]))
        self.external_buying_countries = [tup[0] for tup in self.columns if tup[1] == export_label]
        self.external_selling_countries = [tup[0] for tup in self.index if tup[1] == import_label]
        self.region_households = [tup for tup in self.columns.to_list() if tup[1] == final_demand_label]
        self.adjust_output()

    @classmethod
    def load_mrio_from_filepath(cls, filepath_mrio):
        table = pd.read_csv(filepath_mrio, header=[0, 1], index_col=[0, 1])
        # remove region_sectors with no flows
        zero_output = table.index[table.sum(axis=1) == 0].to_list()
        zero_input = table.columns[table.sum(axis=0) == 0].to_list()
        no_flow_region_sectors = list(set(zero_output) & set(zero_input))
        table.drop(index=no_flow_region_sectors, inplace=True)
        table.drop(columns=no_flow_region_sectors, inplace=True)
        return cls(table)

    def get_total_output_per_region_sectors(self):
        return self.loc[self.region_sectors].sum(axis=1)

    def get_total_input_per_region_sectors(self):
        return self[self.region_sectors].sum()

    def get_region_sectors_with_internal_flows(self, threshold: float = 0):
        tot_outputs = self.get_total_output_per_region_sectors()
        matrix_output = pd.concat([tot_outputs] * len(self.index), axis=1).transpose()
        matrix_output.index = self.index
        tech_coef_matrix = self[self.region_sectors] / matrix_output
        return [tup for tup in self.region_sectors if tech_coef_matrix.loc[tup, tup] > threshold]

    def check_square_structure(self):
        region_sectors_in_columns = [tup for tup in self.columns if tup[1] not in [final_demand_label, export_label]]
        region_sectors_in_rows = [tup for tup in self.index if tup[1] != import_label]
        if region_sectors_in_columns != region_sectors_in_rows:
            logging.info(set(region_sectors_in_columns) - set(region_sectors_in_rows))
            logging.info(set(region_sectors_in_rows) - set(region_sectors_in_columns))
            raise ValueError('Inconsistencies between scope sectors in rows and columns')

    def adjust_output(self):
        total_output = self.get_total_output_per_region_sectors()
        total_input = self.get_total_input_per_region_sectors()
        unbalanced_region_sectors = total_input[total_input > total_output].index.to_list()
        if len(unbalanced_region_sectors) > 0:
            logging.warning(f"There are {len(unbalanced_region_sectors)} region_sectors with more inputs than outputs")
            # TODO very ad-hoc and dirty to accommodate with bad input file
            for region_sector in unbalanced_region_sectors:
                self.loc[region_sector, ('ROW', export_label)] += total_input[region_sector] - total_output[region_sector] + EPSILON

    def get_tech_coef_dict(self, threshold=0):
        """
        returns a dict region_sector = {input_region_sector_1: tech_coef, ...}
        """
        tot_outputs = self.get_total_output_per_region_sectors()
        matrix_output = pd.concat([tot_outputs] * len(self.index), axis=1).transpose()
        matrix_output.index = self.index
        tech_coef_matrix = self[self.region_sectors] / matrix_output
        return {
            '_'.join(region_sector_tuple): {
                '_'.join(key): val
                for key, val in sublist.items()
                if val > threshold
            }
            for region_sector_tuple, sublist in tech_coef_matrix.to_dict().items()
        }



import importlib
import logging
import os
import sys
from datetime import datetime
from pathlib import Path


def check_script_call(arguments: list[str]):
    """
    Usage should be python main.py scope <one optional argument>

    Parameters
    ----------
    arguments

    Returns
    -------

    """
    if len(arguments) < 2:
        raise ValueError('The script takes at least one argument: the scope to be studied. '
                         'Ex. python main.py scope <one optional argument>')
    if len(arguments) > 3:
        raise ValueError('The script does not take more than one optional argument '
                         'Ex. python main.py scope <one optional argument>')
    accepted_optional_arguments: list[str] = [
        'same_transport_network_new_agents',
        'same_agents_new_sc_network',
        'same_sc_network_new_logistic_routes',
        'same_logistic_routes',
        'same_agents_new_transport_network',
        'same_sc_network_new_transport_network',
        'new_agents_same_all'
    ]
    if len(arguments) > 2:
        if arguments[2] not in accepted_optional_arguments:
            raise ValueError("Argument " + arguments[2] + " is not valid.\
                Possible values are: " + ','.join(accepted_optional_arguments))




from setuptools import setup, find_packages

setup(
    name="disruptsc",
    version="1.0.1",
    description="A spatial agent-based model to simulate the dynamics supply chains subject to disruptions",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Celian Colon",
    author_email="celian.colon.2007@polytechnique.org",
    url="https://github.com/ccolon/disruptsc",
    packages=find_packages(),
    license="CC BY-NC-ND 4.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    install_requires=[
        "cpython",
        "geopandas==0.11.1",
        "networkx==2.8.6",
        "PyYAML==6.0"
    ],
)



import pathlib
import sys
#logger = logging.getLogger(__name__)

ROOT_FOLDER = pathlib.Path(__file__).parent.parent
PARAMETER_FOLDER = ROOT_FOLDER / "parameter"
OUTPUT_FOLDER = ROOT_FOLDER / "output"
INPUT_FOLDER = ROOT_FOLDER / "input"
TMP_FOLDER = ROOT_FOLDER / "tmp"

sys.path.insert(1, str(ROOT_FOLDER))
sys.path.insert(1, str(ROOT_FOLDER / "code" / "model"))
sys.path.insert(1, str(ROOT_FOLDER / "code" / "parameters"))
sys.path.insert(1, str(ROOT_FOLDER / "code" / "simualations"))
# if __file__ == "__main__":
#     print(ROOT_FOLDER)

# code_directory = Path(os.path.abspath(__file__)).parent
# project_directory = code_directory.parent
# working_directory = Path(os.getcwd())
# working_directory_parent = working_directory.parent


import logging
from pathlib import Path

import geopandas
import pandas
import pandas as pd

from src.agents.country import Country, Countries
from src.model.basic_functions import rescale_monetary_values
from src.network.mrio import Mrio


def create_countries(filepath_imports: Path, filepath_exports: Path, filepath_transit: Path,
                     transport_nodes: geopandas.GeoDataFrame, present_sectors: list,
                     countries_to_include: list | str = 'all', time_resolution: str = "week",
                     target_units: str = "mUSD", input_units: str = "USD") -> Countries:
    """Create the countries

    Parameters
    ----------
    filepath_imports : string
        path to import table csv

    filepath_exports : string
        path to export table csv

    filepath_transit : string
        path to transit matrix csv

    transport_nodes : pandas.DataFrame
        transport nodes

    present_sectors : list of string
        list which sectors are included. Output of the rescaleFirms functions.

    countries_to_include : list of string or "all"
        List of countries to include. Default to "all", which select all sectors.

    time_resolution : see rescaleMonetaryValues
    target_units : see rescaleMonetaryValues
    input_units : see rescaleMonetaryValues

    Returns
    -------
    list of Countries
    """
    logging.info('Creating countries. Countries included: ' + str(countries_to_include))

    import_table = rescale_monetary_values(
        pd.read_csv(filepath_imports, index_col=0),
        time_resolution=time_resolution,
        target_units=target_units,
        input_units=input_units
    )
    export_table = rescale_monetary_values(
        pd.read_csv(filepath_exports, index_col=0),
        time_resolution=time_resolution,
        target_units=target_units,
        input_units=input_units
    )
    if filepath_transit:
        transit_matrix = rescale_monetary_values(
            pd.read_csv(filepath_transit, index_col=0),
            time_resolution=time_resolution,
            target_units=target_units,
            input_units=input_units
        )
    else:
        # if no transit matrix provide, create a mock one with 0s
        transit_matrix = pd.DataFrame(0, index=import_table.index, columns=import_table.index)
    # entry_point_table = pd.read_csv(filepath_entry_points)

    # Keep only selected countries, if applicable
    if isinstance(countries_to_include, list):
        import_table = import_table.loc[countries_to_include, present_sectors]
        export_table = export_table.loc[countries_to_include, present_sectors]
        transit_matrix = transit_matrix.loc[countries_to_include, countries_to_include]
    elif countries_to_include == 'all':
        import_table = import_table.loc[:, present_sectors]
        export_table = export_table.loc[:, present_sectors]
    else:
        raise ValueError("'countries_to_include' should be a list of string or 'all'")

    logging.info("Total imports per " + time_resolution + " is " +
                 "{:.01f} ".format(import_table.sum().sum()) + target_units)
    logging.info("Total exports per " + time_resolution + " is " +
                 "{:.01f} ".format(export_table.sum().sum()) + target_units)
    logging.info("Total transit per " + time_resolution + " is " +
                 "{:.01f} ".format(transit_matrix.sum().sum()) + target_units)

    country_list = []
    total_imports = import_table.sum().sum()
    for country in import_table.index.tolist():
        cond_country = transport_nodes['special'] == country
        od_point = transport_nodes.loc[cond_country, "id"]
        lon = transport_nodes.geometry.x
        lat = transport_nodes.geometry.y
        if len(od_point) == 0:
            raise ValueError('No od_point found for ' + country)
        elif len(od_point) > 2:
            raise ValueError('More than 1 od_point for ' + country)
        else:
            od_point = od_point.iloc[0]
            lon = lon.iloc[0]
            lat = lat.iloc[0]

        # imports, i.e., sales of countries
        qty_sold = import_table.loc[country, :]
        qty_sold = qty_sold[qty_sold > 0].to_dict()
        supply_importance = sum(qty_sold.values()) / total_imports

        # exports, i.e., purchases from countries
        qty_purchased = export_table.loc[country, :]
        qty_purchased = qty_purchased[qty_purchased > 0].to_dict()

        # transits
        # Note that transit are not given per sector, so, if we only consider a few sector,
        # the full transit flows will still be used
        transit_from = transit_matrix.loc[:, country]
        transit_from = transit_from[transit_from > 0].to_dict()
        transit_to = transit_matrix.loc[country, :]
        transit_to = transit_to[transit_to > 0].to_dict()

        # create the list of Country object
        country_list += [Country(pid=country,
                                 qty_sold=qty_sold,
                                 qty_purchased=qty_purchased,
                                 od_point=od_point,
                                 long=lon,
                                 lat=lat,
                                 transit_from=transit_from,
                                 transit_to=transit_to,
                                 supply_importance=supply_importance
                                 )]
    countries = Countries(country_list)

    logging.info('Country_list created: ' + str([country.pid for country in country_list]))

    return countries


def create_countries_from_mrio(filepath_mrio: Path,
                               transport_nodes: geopandas.GeoDataFrame, time_resolution: str,
                               target_units: str, input_units: str) -> Countries:
    logging.info('Creating countries.')

    # Load mrio
    # mrio = pd.read_csv(filepath_mrio, header=[0, 1], index_col=[0, 1])
    mrio = Mrio.load_mrio_from_filepath(filepath_mrio)

    # Extract countries from MRIO
    buying_countries = mrio.external_buying_countries
    selling_countries = mrio.external_selling_countries
    country_list = list(set(buying_countries) | set(selling_countries))

    # Create country table
    country_table = pd.DataFrame({"pid": country_list})
    logging.info(f"Select {country_table.shape[0]} countries")

    # Extract import, export, and transit matrices
    # importing_region_sectors = [tup for tup in mrio.columns if tup[1] not in ['Exports', 'final_demand']]
    import_table = rescale_monetary_values(
        mrio.loc[(selling_countries, 'Imports'), mrio.region_sectors],
        time_resolution=time_resolution,
        target_units=target_units,
        input_units=input_units
    )
    import_table.index = [tup[0] for tup in import_table.index]
    import_table.columns = ['_'.join(tup) for tup in import_table.columns]
    # exporting_region_sectors = [tup for tup in mrio.index if tup[1] not in ['Imports', 'final_demand']]
    export_table = rescale_monetary_values(
        mrio.loc[mrio.region_sectors, (buying_countries, 'Exports')],
        time_resolution=time_resolution,
        target_units=target_units,
        input_units=input_units
    )
    export_table.columns = [tup[0] for tup in export_table.columns]
    export_table.index = ['_'.join(tup) for tup in export_table.index]
    transit_matrix = rescale_monetary_values(
        mrio.loc[(selling_countries, 'Imports'), (buying_countries, 'Exports')],
        time_resolution=time_resolution,
        target_units=target_units,
        input_units=input_units
    )
    transit_matrix.columns = [tup[0] for tup in transit_matrix.columns]
    transit_matrix.index = [tup[0] for tup in transit_matrix.index]

    logging.info("Total imports per " + time_resolution + " is " +
                 "{:.01f} ".format(import_table.sum().sum()) + target_units)
    logging.info("Total exports per " + time_resolution + " is " +
                 "{:.01f} ".format(export_table.sum().sum()) + target_units)
    logging.info("Total transit per " + time_resolution + " is " +
                 "{:.01f} ".format(transit_matrix.sum().sum()) + target_units)

    total_imports = import_table.sum().sum()
    countries = Countries()
    for country in country_list:
        cond_country = transport_nodes['special'] == country
        od_point = transport_nodes.loc[cond_country, "id"]
        lon = transport_nodes.geometry.x
        lat = transport_nodes.geometry.y
        if len(od_point) == 0:
            raise ValueError('No od_point found for ' + country)
        elif len(od_point) > 2:
            raise ValueError('More than 1 od_point for ' + country)
        else:
            od_point = od_point.iloc[0]
            lon = lon.iloc[0]
            lat = lat.iloc[0]

        # imports, i.e., sales of countries
        if country in selling_countries and total_imports > 0:
            qty_sold = import_table.loc[country, :]
            qty_sold = qty_sold[qty_sold > 0].to_dict()
            supply_importance = sum(qty_sold.values()) / total_imports
        else:
            qty_sold = {}
            supply_importance = 0

        # exports, i.e., purchases from countries
        if country in buying_countries:
            qty_purchased = export_table.loc[:, country]
            qty_purchased = qty_purchased[qty_purchased > 0].to_dict()
        else:
            qty_purchased = {}

        # transits
        # Note that transit are not given per sector, so, if we only consider a few sector,
        # the full transit flows will still be used
        if country in transit_matrix.columns:
            transit_from = transit_matrix.loc[:, country]
            transit_from = transit_from[transit_from > 0].to_dict()
        else:
            transit_from = {}
        if country in transit_matrix.index:
            transit_to = transit_matrix.loc[country, :]
            transit_to = transit_to[transit_to > 0].to_dict()
        else:
            transit_to = {}

        # Populate countries
        countries[country] = Country(pid=country,
                                     qty_sold=qty_sold,
                                     qty_purchased=qty_purchased,
                                     od_point=od_point,
                                     long=lon,
                                     lat=lat,
                                     transit_from=transit_from,
                                     transit_to=transit_to,
                                     supply_importance=supply_importance)
    print(len(countries))
    logging.info('Countries created: ' + str(countries.get_properties('pid')))

    return countries



import random
from typing import TYPE_CHECKING
import logging

import pandas

if TYPE_CHECKING:
    from src.network.sc_network import ScNetwork
    from src.network.transport_network import TransportNetwork
    from src.network.commercial_link import CommercialLink


EPSILON = 1e-6


class Agent(object):
    def __init__(self, agent_type, pid, od_point=0, name=None,
                 long=None, lat=None):
        self.agent_type = agent_type
        self.pid = pid
        self.od_point = od_point
        self.name = name
        self.long = long
        self.lat = lat
        self.usd_per_ton = None

    def id_str(self):
        return f"{self.agent_type} {self.pid} located {self.od_point}".capitalize()

    def receive_shipment_and_pay(self, commercial_link: "CommercialLink", transport_network: "TransportNetwork"):
        """Firm look for shipments in the transport nodes it is located
        It takes those which correspond to the commercial link
        It receives them, thereby removing them from the transport network
        Then it pays the corresponding supplier along the commecial link
        """
        # Look at available shipment
        available_shipments = transport_network._node[self.od_point]['shipments']
        if commercial_link.pid in available_shipments.keys():
            # Identify shipment
            shipment = available_shipments[commercial_link.pid]
            # Get quantity and price
            quantity_delivered = shipment['quantity']
            price = shipment['price']
            # Remove shipment from transport
            transport_network.remove_shipment(commercial_link)
            # Make payment
            commercial_link.payment = quantity_delivered * price
            # If firm, add to inventory
            if self.agent_type == 'firm':
                self.inventory[commercial_link.product] += quantity_delivered

        # If none is available and if there was order, log it
        else:
            if (commercial_link.delivery > 0) and (commercial_link.order > 0):
                logging.info(f"{self.id_str()} - no shipment available for commercial link {commercial_link.pid} "
                             f"({commercial_link.delivery}) of {commercial_link.product})")
            quantity_delivered = 0
            price = 1

        self.update_indicator(quantity_delivered, price, commercial_link)

    def receive_products_and_pay(self, sc_network: "ScNetwork", transport_network: "TransportNetwork",
                                 sectors_no_transport_network: list):
        # reset variable
        self.reset_indicators()

        # for each incoming link, receive product and pay
        # the way differs between service and shipment
        for edge in sc_network.in_edges(self):
            if sc_network[edge[0]][self]['object'].product_type in sectors_no_transport_network:
                self.receive_service_and_pay(sc_network[edge[0]][self]['object'])
            else:
                self.receive_shipment_and_pay(sc_network[edge[0]][self]['object'], transport_network)

    def receive_service_and_pay(self, commercial_link):
        # Always available, same price
        quantity_delivered = commercial_link.delivery
        commercial_link.payment = quantity_delivered * commercial_link.price
        # Update indicator
        self.update_indicator(quantity_delivered, commercial_link.price, commercial_link)

    def update_indicator(self, quantity_delivered: float, price: float, commercial_link: "CommercialLink"):
        """When receiving product, agents update some internal variables
        """
        # Log if quantity received differs from what it was supposed to be
        if abs(commercial_link.delivery - quantity_delivered) > EPSILON:
            logging.debug(
                f"{self.id_str()} - Quantity delivered by {commercial_link.supplier_id} is {quantity_delivered};"
                f" It was supposed to be {commercial_link.delivery}.")

    def choose_initial_routes(self, sc_network: "ScNetwork", transport_network: "TransportNetwork",
                              capacity_constraint: bool,
                              transport_cost_noise_level: float, monetary_unit_flow: str):
        for edge in sc_network.out_edges(self):
            if edge[1].od_point == -1:  # we do not create route for service firms if explicit_service_firms = False
                continue
            else:
                # Get the id of the origin and destination node
                origin_node = self.od_point
                destination_node = edge[1].od_point
                # Choose the route and the corresponding mode
                route = self.choose_route(
                    transport_network=transport_network,
                    origin_node=origin_node,
                    destination_node=destination_node,
                    capacity_constraint=capacity_constraint,
                    transport_cost_noise_level=transport_cost_noise_level
                )
                # Store it into commercial link object
                sc_network[self][edge[1]]['object'].store_route_information(
                    route=route,
                    main_or_alternative="main"
                )

                if capacity_constraint:
                    self.update_transport_load(edge, monetary_unit_flow, route, sc_network, transport_network,
                                               capacity_constraint)

    def get_transport_cond(self, edge, transport_modes):
        # Define the type of transport mode to use by looking in the transport_mode table
        if self.agent_type == 'firm':
            cond_from = (transport_modes['from'] == "domestic")
        elif self.agent_type == 'country':
            cond_from = (transport_modes['from'] == self.pid)
        else:
            raise ValueError("'self' must be a Firm or a Country")
        if edge[1].agent_type in ['firm', 'household']:  # see what is the other end
            cond_to = (transport_modes['to'] == "domestic")
        elif edge[1].agent_type == 'country':
            cond_to = (transport_modes['to'] == edge[1].pid)
        else:
            raise ValueError("'edge[1]' must be a Firm or a Country")
            # we have not implemented a "sector" condition
        return cond_from, cond_to

    def update_transport_load(self, edge, monetary_unit_flow, route, sc_network, transport_network, capacity_constraint):
        # Update the "current load" on the transport network
        # if current_load exceed burden, then add burden to the weight
        new_load_in_usd = sc_network[self][edge[1]]['object'].order
        new_load_in_tons = Agent.transformUSD_to_tons(new_load_in_usd, monetary_unit_flow, self.usd_per_ton)
        transport_network.update_load_on_route(route, new_load_in_tons, capacity_constraint)

    def choose_route(self, transport_network: "TransportNetwork", origin_node: int, destination_node: int,
                     capacity_constraint: bool, transport_cost_noise_level: float):
        """
        The agent choose the delivery route

        The only way re-implemented (vs. Cambodian version) ist that any mode can be chosen

        Keeping here the comments of the Cambodian version
        If the simple case in which there is only one accepted_logistics_modes
        (as defined by the main parameter logistic_modes)
        then it is simply the shortest_route using the appropriate weigh

        If there are several accepted_logistics_modes, then the agent will investigate different route,
        one per accepted_logistics_mode. They will then pick one, with a certain probability taking into account the
        weight This more complex mode is used when, according to the capacity and cost data, all the exports or
        imports are using one route, whereas in the data, we observe still some flows using another mode of
        transport. So we use this to "force" some flow to take the other routes.
        """
        if capacity_constraint:
            weight_considered = "capacity_weight"
        else:
            weight_considered = "weight"
        route = transport_network.provide_shortest_route(origin_node,
                                                         destination_node,
                                                         route_weight=weight_considered,
                                                         noise_level=transport_cost_noise_level)
        # if route is None:
        #     raise ValueError(f"Agent {self.pid} - No route found from {origin_node} to {destination_node}")
        # else:
        return route

    @staticmethod
    def check_route_availability(commercial_link, transport_network, which_route='main'):
        """
        Look at the main or alternative route
        at check all edges and nodes in the route
        if one is marked as disrupted, then the whole route is marked as disrupted
        """

        if which_route == 'main':
            route_to_check = commercial_link.route
        elif which_route == 'alternative':
            route_to_check = commercial_link.alternative_route
        else:
            raise KeyError('Wrong value for parameter which_route, admissible values are main and alternative')

        res = 'available'
        for route_segment in route_to_check:
            if len(route_segment) == 2:
                if transport_network[route_segment[0]][route_segment[1]]['disruption_duration'] > 0:
                    res = 'disrupted'
                    break
            if len(route_segment) == 1:
                if transport_network._node[route_segment[0]]['disruption_duration'] > 0:
                    res = 'disrupted'
                    break
        return res

    @staticmethod
    def transformUSD_to_tons(monetary_flow, monetary_unit, usd_per_ton):
        if usd_per_ton == 0:
            return 0
        else:
            # Load monetary units
            monetary_unit_factor = {
                "mUSD": 1e6,
                "kUSD": 1e3,
                "USD": 1
            }
            factor = monetary_unit_factor[monetary_unit]
            return monetary_flow / (usd_per_ton / factor)

    def reset_indicators(self):
        pass


class Agents(dict):
    def __init__(self, agent_list=None):
        super().__init__()
        if agent_list is not None:
            for agent in agent_list:
                self[agent.pid] = agent

    def __setitem__(self, key, value):
        if not isinstance(value, Agent):
            raise KeyError("Value must be an Agent")
        if not hasattr(value, 'pid'):
            raise ValueError("Value must have a 'pid' attribute")
        super().__setitem__(key, value)

    def add(self, agent: Agent):
        if not hasattr(agent, 'pid'):
            raise ValueError("Object must have a 'pid' attribute")
        self[agent.pid] = agent

    def __repr__(self):
        return f"PidDict({super().__repr__()})"

    def sum(self, property_name):
        total = 0
        for agent in self.values():
            if hasattr(agent, property_name):
                total += getattr(agent, property_name)
            else:
                raise AttributeError(f"Agent does not have the property '{property_name}'")
        return total

    def mean(self, property_name):
        total = 0
        count = 0
        for agent in self.values():
            if hasattr(agent, property_name):
                total += getattr(agent, property_name)
                count += 1
            else:
                raise AttributeError(f"Agent does not have the property '{property_name}'")
        if count == 0:
            raise ValueError(f"No agents with the property '{property_name}' found.")
        return total / count

    def get_properties(self, property_name):
        return {pid: getattr(agent, property_name) for pid, agent in self.items()}

    def send_purchase_orders(self, sc_network: "ScNetwork"):
        for agent in self.values():
            agent.send_purchase_orders(sc_network)

    def deliver(self, sc_network: "ScNetwork", transport_network: "TransportNetwork",
                sectors_no_transport_network: list, rationing_mode: str, capacity_constraint: bool,
                monetary_units_in_model: str, cost_repercussion_mode: str, price_increase_threshold: float,
                transport_cost_noise_level: float):
        for agent in self.values():
            agent.deliver_products(sc_network, transport_network,
                                   sectors_no_transport_network=sectors_no_transport_network,
                                   rationing_mode=rationing_mode, monetary_units_in_model=monetary_units_in_model,
                                   cost_repercussion_mode=cost_repercussion_mode,
                                   price_increase_threshold=price_increase_threshold,
                                   capacity_constraint=capacity_constraint,
                                   transport_cost_noise_level=transport_cost_noise_level)

    def receive_products(self, sc_network: "ScNetwork", transport_network: "TransportNetwork",
                         sectors_no_transport_network: list):
        for agent in self.values():
            agent.receive_products_and_pay(sc_network, transport_network, sectors_no_transport_network)


def determine_nb_suppliers(nb_suppliers_per_input: float, max_nb_of_suppliers=None):
    """Draw 1 or 2 depending on the 'nb_suppliers_per_input' parameters

    nb_suppliers_per_input is a float number between 1 and 2

    max_nb_of_suppliers: maximum value not to exceed
    """
    if (nb_suppliers_per_input < 1) or (nb_suppliers_per_input > 2):
        raise ValueError("'nb_suppliers_per_input' should be between 1 and 2")

    if nb_suppliers_per_input == 1:
        nb_suppliers = 1

    elif nb_suppliers_per_input == 2:
        nb_suppliers = 2

    else:
        if random.uniform(0, 1) < nb_suppliers_per_input - 1:
            nb_suppliers = 2
        else:
            nb_suppliers = 1

    if max_nb_of_suppliers:
        nb_suppliers = min(nb_suppliers, max_nb_of_suppliers)

    return nb_suppliers
